package org.dimdev.dimdoors.api.util.math;

import java.util.Map;
import net.minecraft.class_1297;
import net.minecraft.class_2350;
import net.minecraft.class_2379;
import net.minecraft.class_243;
import net.minecraft.class_5819;
import org.dimdev.dimdoors.api.util.math.TransformationMatrix3d;

public final class MathUtil {
	public static <T> T weightedRandom(Map<T, Float> weights) {
		if (weights.size() == 0) return null;
		int totalWeight = 0;
		for (float weight : weights.values()) {
			totalWeight += weight;
		}
		class_5819 random = class_5819.method_43047();
		float f = random.method_43057() * totalWeight;
		for (Map.Entry<T, Float> e : weights.entrySet()) {
			f -= e.getValue();
			if (f < 0) return e.getKey();
		}
		return null;
	}

	public static class_2379 eulerAngle(class_243 direction, class_243 upwards) {
		float pitch = pitch(direction);
		float yaw = yaw(direction);
		upwards = TransformationMatrix3d.builder().rotate(new class_2379(pitch, yaw, 0)).buildReverse().transform(upwards);
		float roll = (float) Math.toDegrees(-Math.atan2(upwards.field_1352, upwards.field_1351));

		return new class_2379(pitch, yaw, roll);
	}

	public static class_2379 entityEulerAngle(class_1297 entity) {
		return new class_2379(entity.method_36455(), entity.method_36454(), 0);
	}

	public static float yaw(class_243 vector) {
		return (float) Math.toDegrees(-Math.atan2(vector.field_1352, vector.field_1350));
	}

	public static float pitch(class_243 vector) {
		return (float) Math.toDegrees(Math.asin(-vector.field_1351));
	}

	public static class_2379 directionEulerAngle(class_2350 direction) {
		switch (direction) {
			case field_11033:
				return EulerAngleDirection.DOWN.getAngle();
			case field_11036:
				return EulerAngleDirection.UP.getAngle();
			case field_11043:
				return EulerAngleDirection.NORTH.getAngle();
			case field_11035:
				return EulerAngleDirection.SOUTH.getAngle();
			case field_11039:
				return EulerAngleDirection.WEST.getAngle();
			case field_11034:
			default:
				return EulerAngleDirection.EAST.getAngle();
		}
	}

	public enum EulerAngleDirection {
		DOWN(new class_2379(90, 0, 0)),
		UP(new class_2379(-90, 0, 0)),
		NORTH(new class_2379(0, -180, 0)),
		SOUTH(new class_2379(0, 0, 0)),
		WEST(new class_2379(0, 90, 0)),
		EAST(new class_2379(0, -90, 0));



		private final class_2379 angle;

		EulerAngleDirection(class_2379 angle) {
			this.angle = angle;
		}

		public class_2379 getAngle() {
			return angle;
		}
	}
}
