package org.dimdev.dimdoors.api.util.math;

import net.minecraft.class_2379;
import net.minecraft.class_243;

public class TransformationMatrix3d extends TransformationMatrixdImpl<TransformationMatrix3d> {
	public TransformationMatrix3d(double[][] matrix) {
		super(matrix);
	}

	public TransformationMatrix3d(AbstractMatrixd<? extends AbstractMatrixd<?>> matrix) {
		super(matrix);
	}

	public TransformationMatrix3d(Vectord... vectors) {
		super(vectors);
	}

	public static TransformationMatrix3d identity() {
		return new TransformationMatrix3d(Matrixd.identity(4, 4));
	}

	@Override
	public TransformationMatrix3d construct(double[][] matrix) {
		return new TransformationMatrix3d(matrix);
	}

	@Override
	public TransformationMatrix3d construct(AbstractMatrixd<? extends AbstractMatrixd<?>> matrixd) {
		return new TransformationMatrix3d(matrixd);
	}

	@Override
	public TransformationMatrix3d construct(Vectord... vectors) {
		return new TransformationMatrix3d(vectors);
	}

	public class_243 transform(class_243 vector) {
		Vectord vec = transform(new Vectord(vector.field_1352, vector.field_1351, vector.field_1350));
		return new class_243(vec.get(0), vec.get(1), vec.get(2));
	}

	// Should only be called on pure rotation matrices, behaviour undefined otherwise.
	public class_2379 transform(class_2379 angle) {
		TransformationMatrix3d rotator = TransformationMatrix3d.builder().rotate(angle).build();
		// angle vector representation
		class_243 direction = rotator.transform(new class_243(0, 0, 1));
		class_243 upwards = rotator.transform(new class_243(0, 1, 0));

		direction = transform(direction);
		upwards = transform(upwards);

		return MathUtil.eulerAngle(direction, upwards);
	}

	public static TransformationMatrix3dBuilder builder() {
		return new TransformationMatrix3dBuilder(identity());
	}

	public static class TransformationMatrix3dBuilder extends TransformationMatrixdBuilderImpl<TransformationMatrix3dBuilder, TransformationMatrix3d> {
		protected TransformationMatrix3dBuilder(TransformationMatrix3d instance) {
			super(3, instance);
		}

		@Override
		public TransformationMatrix3dBuilder getSelf() {
			return this;
		}

		public TransformationMatrix3dBuilder translate(class_243 vector) {
			return translate(new Vectord(vector.field_1352, vector.field_1351, vector.field_1350));
		}

		public TransformationMatrix3dBuilder inverseTranslate(class_243 vector) {
			return translate(new Vectord(-vector.field_1352, -vector.field_1351, -vector.field_1350));
		}


		public TransformationMatrix3dBuilder rotateX(double angle) {
			return rotateAroundBasePlane(angle, 1, 2);
		}

		public TransformationMatrix3dBuilder rotateY(double angle) {
			return rotateAroundBasePlane(angle, 2, 0);
		}

		public TransformationMatrix3dBuilder rotateZ(double angle) {
			return rotateAroundBasePlane(angle, 0, 1);
		}

		public TransformationMatrix3dBuilder rotate(class_2379 angle) {
			return this.rotateZ(Math.toRadians(angle.method_10258())) // roll
					.rotateX(Math.toRadians(angle.method_10256())) // pitch
					.rotateY(Math.toRadians(-angle.method_10257())); // yaw
		}

		public TransformationMatrix3dBuilder inverseRotate(class_2379 angle) {
			return this.rotateZ(-Math.toRadians(angle.method_10258())) // roll
					.rotateX(-Math.toRadians(angle.method_10256())) // pitch
					.rotateY(-Math.toRadians(-angle.method_10257())); // yaw
		}
	}
}
