package org.dimdev.dimdoors.block.door;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import net.fabricmc.api.EnvType;
import net.fabricmc.fabric.api.blockrenderlayer.v1.BlockRenderLayerMap;
import net.fabricmc.fabric.api.event.registry.RegistryEntryAddedCallback;
import net.fabricmc.fabric.api.object.builder.v1.block.FabricBlockSettings;
import net.fabricmc.loader.api.FabricLoader;
import net.minecraft.block.*;
import net.minecraft.class_1921;
import net.minecraft.class_2248;
import net.minecraft.class_2323;
import net.minecraft.class_2378;
import net.minecraft.class_2533;
import net.minecraft.class_2588;
import net.minecraft.class_2680;
import net.minecraft.class_2689;
import net.minecraft.class_2769;
import net.minecraft.class_2960;
import net.minecraft.class_4970;
import net.minecraft.class_5250;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.block.entity.ModBlockEntityTypes;
import org.dimdev.dimdoors.item.DimensionalDoorItemRegistrar;
import org.dimdev.dimdoors.listener.BlockRegistryEntryAddedListener;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.function.BiFunction;

public class DimensionalDoorBlockRegistrar {
	private static final String PREFIX = "block_ag_dim_";

	private final class_2378<class_2248> registry;
	private final DimensionalDoorItemRegistrar itemRegistrar;

	private final BiMap<class_2960, class_2960> mappedDoorBlocks = HashBiMap.create();

	public DimensionalDoorBlockRegistrar(class_2378<class_2248> registry, DimensionalDoorItemRegistrar itemRegistrar) {
		this.registry = registry;
		this.itemRegistrar = itemRegistrar;

		init();
		RegistryEntryAddedCallback.event(registry).register(new BlockRegistryEntryAddedListener(this));
	}

	private void init() {
		new ArrayList<>(registry.method_29722())
				.forEach(entry -> handleEntry(entry.getKey().method_29177(), entry.getValue()));
	}

	public void handleEntry(class_2960 identifier, class_2248 original) {
		if (DimensionalDoorsInitializer.getConfig().getDoorsConfig().isAllowed(identifier)) {
			if (!(original instanceof DimensionalDoorBlock) && original instanceof class_2323) {
				register(identifier, original, DimensionalDoorBlockRegistrar::createAutoGenDimensionalDoorBlock);
			} else if (!(original instanceof DimensionalTrapdoorBlock) && original instanceof class_2533) {
				register(identifier, original, DimensionalDoorBlockRegistrar::createAutoGenDimensionalTrapdoorBlock);
			}
		}
	}

	private void register(class_2960 identifier, class_2248 original, BiFunction<class_4970.class_2251, class_2248, ? extends class_2248> constructor) {
		class_2960 gennedId = new class_2960("dimdoors", PREFIX + identifier.method_12836() + "_" + identifier.method_12832());
		class_2248 dimBlock = class_2378.method_10230(registry, gennedId, constructor.apply(FabricBlockSettings.method_9630(original), original));
		ModBlockEntityTypes.ENTRANCE_RIFT.addBlock(dimBlock);
		mappedDoorBlocks.put(gennedId, identifier);
		itemRegistrar.notifyBlockMapped(original, dimBlock);

		if (FabricLoader.getInstance().getEnvironmentType() == EnvType.CLIENT) {
			putCutout(dimBlock);
		}
	}

	private void putCutout(class_2248 original) {
		BlockRenderLayerMap.INSTANCE.putBlock(original, class_1921.method_23581());
	}

	public class_2960 get(class_2960 identifier) {
		return mappedDoorBlocks.get(identifier);
	}

	public boolean isMapped(class_2960 identifier) {
		return mappedDoorBlocks.containsKey(identifier);
	}

	private static <T extends Comparable<T>> class_2680 transferProperty(class_2680 from, class_2680 to, class_2769<T> property) {
		return to.method_11657(property, from.method_11654(property));
	}

	private static AutoGenDimensionalDoorBlock createAutoGenDimensionalDoorBlock(class_4970.class_2251 settings, class_2248 originalBlock) {
		return new AutoGenDimensionalDoorBlock(settings, originalBlock) {
			@Override
			protected void method_9515(class_2689.class_2690<class_2248, class_2680> builder) {
				appendPropertiesOverride(builder, originalBlock, WATERLOGGED);
			}
		};
	}

	private static AutoGenDimensionalTrapdoorBlock createAutoGenDimensionalTrapdoorBlock(class_4970.class_2251 settings, class_2248 originalBlock) {
		return new AutoGenDimensionalTrapdoorBlock(settings, originalBlock) {
			@Override
			protected void method_9515(class_2689.class_2690<class_2248, class_2680> builder) {
				appendPropertiesOverride(builder, originalBlock, field_11626);
			}
		};
	}

	private static void appendPropertiesOverride(class_2689.class_2690<class_2248, class_2680> builder, class_2248 originalBlock, class_2769<?>... requiredProperties) {
		HashSet<class_2769<?>> properties = new HashSet<>(originalBlock.method_9595().method_11659());
		properties.addAll(List.of(requiredProperties));
		builder.method_11667(properties.toArray(new class_2769[0]));
	}

	private static class AutoGenDimensionalDoorBlock extends DimensionalDoorBlock {
		private final class_2248 originalBlock;

		public AutoGenDimensionalDoorBlock(class_2251 settings, class_2248 originalBlock) {
			super(settings);
			this.originalBlock = originalBlock;

			class_2680 state = this.method_9595().method_11664();
			class_2680 originalState = originalBlock.method_9564();
			for (class_2769<?> property : originalBlock.method_9564().method_28501()) {
				state = transferProperty(originalState, state, property);
			}
			method_9590(state.method_11657(WATERLOGGED, false));
		}

		@Override
		protected void method_9515(class_2689.class_2690<class_2248, class_2680> builder) {
			// This method has to be defined in an anonymous inner class,
			// 		since Block#appendProperties is run before originalBlock can be set.
			throw new RuntimeException("AutoGenDimensionalDoorBlock should be instantiated as anonymous inner class overriding appendProperties!");
		}

		@Override
		public class_5250 method_9518() {
			return class_5250.method_43477(new class_2588("dimdoors.autogen_block_prefix")).method_10852(originalBlock.method_9518());
		}
	}

	private static class AutoGenDimensionalTrapdoorBlock extends DimensionalTrapdoorBlock {
		private final class_2248 originalBlock;

		public AutoGenDimensionalTrapdoorBlock(class_2251 settings, class_2248 originalBlock) {
			super(settings);
			this.originalBlock = originalBlock;

			class_2680 state = this.method_9595().method_11664();
			class_2680 originalState = originalBlock.method_9564();
			for (class_2769<?> property : originalBlock.method_9564().method_28501()) {
				state = transferProperty(originalState, state, property);
			}
			method_9590(state.method_11657(field_11626, false));
		}

		@Override
		protected void method_9515(class_2689.class_2690<class_2248, class_2680> builder) {
			// This method has to be defined in an anonymous inner class,
			// 		since Block#appendProperties is run before originalBlock can be set.
			throw new RuntimeException("AutoGenDimensionalTrapdoorBlock should be instantiated as anonymous inner class overriding appendProperties!");
		}

		@Override
		public class_5250 method_9518() {
			return class_5250.method_43477(new class_2588("dimdoors.autogen_block_prefix")).method_10852(originalBlock.method_9518());
		}
	}
}
