package org.dimdev.dimdoors.client.util;

import com.google.common.base.Preconditions;
import me.shedaniel.clothconfig2.api.ConfigBuilder;
import me.shedaniel.clothconfig2.api.ConfigCategory;
import me.shedaniel.clothconfig2.api.ConfigEntryBuilder;
import me.shedaniel.clothconfig2.impl.builders.BooleanToggleBuilder;
import me.shedaniel.clothconfig2.impl.builders.DoubleFieldBuilder;
import me.shedaniel.clothconfig2.impl.builders.IntFieldBuilder;
import me.shedaniel.clothconfig2.impl.builders.IntListBuilder;
import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_2561;
import net.minecraft.class_2588;
import net.minecraft.class_437;
import net.minecraft.class_5250;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;

@Environment(EnvType.CLIENT)
public class ScreenGenerator {
	public static class_437 create(class_437 parent, Object config, Runnable saveAction) {
		Class<?> configClass = config.getClass();
		Preconditions.checkArgument(configClass.isAnnotationPresent(Title.class));
		Preconditions.checkNotNull(saveAction);
		ConfigBuilder configBuilder = ConfigBuilder.create();
		ConfigEntryBuilder entryBuilder = configBuilder.entryBuilder();
		configBuilder.setTitle(class_5250.method_43477(new class_2588(configClass.getAnnotation(Title.class).value())));
		configBuilder.setSavingRunnable(saveAction);
		configBuilder.setParentScreen(parent);
		for (Field field : configClass.getDeclaredFields()) {
			if (!field.isAnnotationPresent(Category.class)
					|| Modifier.isStatic(field.getModifiers())
					|| !Modifier.isFinal(field.getModifiers())) {
				continue;
			}

			ConfigCategory category = configBuilder.getOrCreateCategory(class_5250.method_43477(new class_2588(configClass.getName().toLowerCase() + ":" + field.getName())));
			Object value;
			try {
				if (field.isAnnotationPresent(Private.class)) {
					field.setAccessible(true);
				}
				value = field.get(config);
			} catch (IllegalAccessException e) {
				throw new AssertionError(e);
			}

			Class<?> valueClass = value.getClass();
			for (Field innerField : valueClass.getFields()) {
				if (!field.isAnnotationPresent(Expose.class)
						|| Modifier.isStatic(field.getModifiers())
						|| Modifier.isFinal(field.getModifiers())) {
					continue;
				}

				Optional<class_2561[]> tooltipSupplier = Optional.of(Optional.of(innerField)
						.filter(f -> f.isAnnotationPresent(Tooltips.class))
						.map(f -> f.getAnnotation(Tooltips.class))
						.map(Tooltips::value)
						.map(Arrays::stream)
						.orElse(Stream.empty())
						.map(tooltip -> {
							if (tooltip.absolute()) return class_2561.method_30163(tooltip.value());
							else return class_5250.method_43477(new class_2588(valueClass.getName().toLowerCase() + "." + innerField.getName() + ":" + tooltip.value()));
						})
						.map(text -> (class_2561) text)
						.toArray(class_2561[]::new))
						.filter(texts -> texts.length > 0);

				try {
					if (innerField.getType() == boolean.class) {
						BooleanToggleBuilder builder = entryBuilder.startBooleanToggle(class_5250.method_43477(new class_2588(valueClass.getName().toLowerCase() + ":" + innerField.getName())), innerField.getBoolean(value))
								.setTooltip(tooltipSupplier)
								.setSaveConsumer(bool -> {
									try {
										innerField.setBoolean(value, bool);
									} catch (IllegalAccessException e) {
										throw new AssertionError();
									}
								});
						builder.requireRestart(field.isAnnotationPresent(RequiresRestart.class));
						category.addEntry(builder.build());
					} else if (innerField.getType() == int.class) {
						IntFieldBuilder builder = entryBuilder.startIntField(class_5250.method_43477(new class_2588(valueClass.getName().toLowerCase() + ":" + innerField.getName())), innerField.getInt(value))
								.setTooltip(tooltipSupplier)
								.setSaveConsumer(i -> {
									try {
										innerField.setInt(value, i);
									} catch (IllegalAccessException e) {
										throw new AssertionError();
									}
								});
						builder.requireRestart(field.isAnnotationPresent(RequiresRestart.class));
						category.addEntry(builder.build());
					} else if (innerField.getType() == double.class) {
						DoubleFieldBuilder builder = entryBuilder.startDoubleField(class_5250.method_43477(new class_2588(valueClass.getName().toLowerCase() + ":" + innerField.getName())), innerField.getDouble(value))
								.setTooltip(tooltipSupplier)
								.setSaveConsumer(d -> {
									try {
										innerField.setDouble(value, d);
									} catch (IllegalAccessException e) {
										throw new AssertionError();
									}
								});
						builder.requireRestart(field.isAnnotationPresent(RequiresRestart.class));
						category.addEntry(builder.build());
					} else if (innerField.isAnnotationPresent(IntSet.class)) {
						IntListBuilder builder = entryBuilder.startIntList(class_5250.method_43477(new class_2588(valueClass.getName().toLowerCase() + ":" + innerField.getName())), new ArrayList<>(((Set<Integer>) innerField.get(value))))
								.setTooltip(tooltipSupplier)
								.setSaveConsumer(set -> {
									try {
										innerField.set(value, set);
									} catch (IllegalAccessException e) {
										throw new AssertionError();
									}
								});
						builder.requireRestart(field.isAnnotationPresent(RequiresRestart.class));
						category.addEntry(builder.build());
					} else {
						throw new AssertionError();
					}
				} catch (IllegalAccessException e) {
					throw new AssertionError("Invalid type " + innerField.getType().getName(), e);
				}
			}
		}
		return configBuilder.build();
	}
}
