package org.dimdev.dimdoors.datagen;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonObject;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.api.util.ResourceUtil;
import org.dimdev.dimdoors.block.ModBlocks;
import org.dimdev.dimdoors.world.decay.DecayPredicate;
import org.dimdev.dimdoors.world.decay.DecayProcessor;
import org.dimdev.dimdoors.world.decay.predicates.SimpleDecayPredicate;
import org.dimdev.dimdoors.world.decay.processors.SelfDecayProcessor;
import org.dimdev.dimdoors.world.decay.processors.SimpleDecayProcesor;
import java.io.IOException;
import java.nio.file.Path;
import java.util.function.BiConsumer;
import net.minecraft.class_2248;
import net.minecraft.class_2403;
import net.minecraft.class_2405;
import net.minecraft.class_2487;
import net.minecraft.class_2960;
import net.minecraft.class_7403;

import static net.minecraft.class_2246.*;

public class LimboDecayProvider implements class_2405 {
    private static final Logger LOGGER = LogManager.getLogger();
    private static final Gson GSON = (new GsonBuilder()).setPrettyPrinting().disableHtmlEscaping().create();

    private final class_2403 generator;

    public LimboDecayProvider(class_2403 generator) {
        this.generator  = generator;
    }

    @Override
    public void method_10319(class_7403 cache) throws IOException {
        Path path = this.generator.method_10313();

        BiConsumer<class_2960, JsonObject> consumer = (identifier, json)  -> {
            Path outputPath = getOutput(path, identifier);

            try {
                class_2405.method_10320(cache, json, outputPath);
            } catch (IOException var6) {
                LOGGER.error("Couldn't save decay pattern {}", outputPath, var6);
        }
        };

		generatePatterns(consumer);
    }

	protected void generatePatterns(BiConsumer<class_2960, JsonObject> consumer) {
		createSimplePattern(new class_2960("dimdoors:stone"), field_10340, field_10445).run(consumer);
		createSimplePattern(new class_2960("dimdoors:cobblestone"), field_10445, field_10255).run(consumer);
		createSimplePattern(new class_2960("dimdoors:gravel"), field_10255, field_10102).run(consumer);
		turnIntoSelf(new class_2960("dimdoors:sand"), field_10102).run(consumer);
		createSimplePattern(new class_2960("dimdoors:bedrock"), field_9987, ModBlocks.SOLID_STATIC);
		createSimplePattern(new class_2960("dimdoors:glass"), field_10033, field_10102).run(consumer);
		createSimplePattern(new class_2960("dimdoors:grass_block"), field_10219, field_10566).run(consumer);
		createSimplePattern(new class_2960("dimdoors:dirt"), field_10566, field_10102).run(consumer);
		createSimplePattern(new class_2960("dimdoors:redstone_block"), field_10002, field_10080).run(consumer);
		createSimplePattern(new class_2960("dimdoors:redstone_ore"), field_10080, field_10340).run(consumer);
		createSimplePattern(new class_2960("dimdoors:emerald_block"), field_10234, field_10013).run(consumer);
		createSimplePattern(new class_2960("dimdoors:emerald_ore"), field_10013, field_10340).run(consumer);
		createSimplePattern(new class_2960("dimdoors:coal_block"), field_10381, field_10418).run(consumer);
		createSimplePattern(new class_2960("dimdoors:coal_ore"), field_10418, field_10340).run(consumer);
		createSimplePattern(new class_2960("dimdoors:iron_block"), field_10085, field_10212).run(consumer);
		createSimplePattern(new class_2960("dimdoors:iron_ore"), field_10212, field_10340).run(consumer);
		createSimplePattern(new class_2960("dimdoors:lapis_block"), field_10441, field_10090).run(consumer);
		createSimplePattern(new class_2960("dimdoors:lapis_ore"), field_10090, field_10340).run(consumer);
		createSimplePattern(new class_2960("dimdoors:gold_block"), field_10205, field_10571).run(consumer);
		createSimplePattern(new class_2960("dimdoors:gold_ore"), field_10571, field_10340).run(consumer);
		createSimplePattern(new class_2960("dimdoors:sandstone"), field_9979, field_10102).run(consumer);
		createSimplePattern(new class_2960("dimdoors:end_stone_bricks"), field_10462, field_10471).run(consumer);
		createSimplePattern(new class_2960("dimdoors:dirt_path"), field_10194, field_10566).run(consumer);
		createSimplePattern(new class_2960("dimdoors:polished_granite"), field_10289, field_10474).run(consumer);
		createSimplePattern(new class_2960("dimdoors:polished_andesite"), field_10093, field_10115).run(consumer);
		createSimplePattern(new class_2960("dimdoors:andesite"), field_10115, field_10508).run(consumer);
		createSimplePattern(new class_2960("dimdoors:polished_diorite"), field_10346, field_10508).run(consumer);
		createSimplePattern(new class_2960("dimdoors:granite"), field_10474, field_10508).run(consumer);
		createSimplePattern(new class_2960("dimdoors:diorite"), field_10508, field_10445).run(consumer);
		createSimplePattern(new class_2960("dimdoors:polished_blackstone"), field_23873, field_23869).run(consumer);
		createSimplePattern(new class_2960("dimdoors:blackstone"), field_23869, field_10445).run(consumer);
		createSimplePattern(new class_2960("dimdoors:podzol"), field_10520, field_10566).run(consumer);
		createSimplePattern(new class_2960("dimdoors:farmland"), field_10362, field_10566).run(consumer);
		createSimplePattern(new class_2960("dimdoors:stone_bricks"), field_10056, field_10416).run(consumer);
		createSimplePattern(new class_2960("dimdoors:cracked_stone_bricks"), field_10416, field_10508).run(consumer);
		createSimplePattern(new class_2960("dimdoors:end_stone"), field_10471, field_9979).run(consumer);
		createSimplePattern(new class_2960("dimdoors:oak_log"), field_10431, field_10161).run(consumer);
		createSimplePattern(new class_2960("dimdoors:birch_log"), field_10511, field_10148).run(consumer);
		createSimplePattern(new class_2960("dimdoors:spruce_log"), field_10037, field_9975).run(consumer);
		createSimplePattern(new class_2960("dimdoors:jungle_log"), field_10306, field_10334).run(consumer);
		createSimplePattern(new class_2960("dimdoors:acacia_log"), field_10533, field_10218).run(consumer);
		createSimplePattern(new class_2960("dimdoors:dark_oak_log"), field_10010, field_10075).run(consumer);
		createSimplePattern(new class_2960("dimdoors:oak_wood"), field_10126, field_10431).run(consumer);
		createSimplePattern(new class_2960("dimdoors:birch_wood"), field_10307, field_10511).run(consumer);
		createSimplePattern(new class_2960("dimdoors:spruce_wood"), field_10155, field_10037).run(consumer);
		createSimplePattern(new class_2960("dimdoors:jungle_wood"), field_10303, field_10306).run(consumer);
		createSimplePattern(new class_2960("dimdoors:acacia_wood"), field_9999, field_10533).run(consumer);
		createSimplePattern(new class_2960("dimdoors:dark_oak_wood"), field_10178, field_10010).run(consumer);
	}

	private DecayPatternData turnIntoSelf(class_2960 identifier, class_2248 before) {
        return new DecayPatternData(identifier, SimpleDecayPredicate.builder().block(before).create(), SelfDecayProcessor.instance());
    }


    @Override
    public String method_10321() {
        return "Limbo Decay";
    }

    private static Path getOutput(Path rootOutput, class_2960 lootTableId) {
        return rootOutput.resolve("data/" + lootTableId.method_12836() + "/decay_patterns/" + lootTableId.method_12832() + ".json");
    }

    public DecayPatternData createSimplePattern(class_2960 id, class_2248 before, class_2248 after) {
        return new DecayPatternData(id, SimpleDecayPredicate.builder().block(before).create(), SimpleDecayProcesor.builder().block(after).entropy(1).create());
    }

    public static class DecayPatternData {
        private class_2960 id;
        private DecayPredicate predicate;
        private DecayProcessor processor;

        public DecayPatternData(class_2960 id, DecayPredicate predicate, DecayProcessor processor) {
            this.id = id;
            this.predicate = predicate;
            this.processor = processor;
        }

        public void run(BiConsumer<class_2960, JsonObject> consumer) {
            JsonObject object = new JsonObject();
            object.add("predicate", ResourceUtil.NBT_TO_JSON.apply(predicate.toNbt(new class_2487())));
            object.add("processor", ResourceUtil.NBT_TO_JSON.apply(processor.toNbt(new class_2487())));

            consumer.accept(id, object);
        }
    }
}
