package org.dimdev.dimdoors.item;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.fabricmc.fabric.api.client.rendering.v1.BuiltinItemRendererRegistry;
import net.fabricmc.fabric.api.event.registry.RegistryEntryAddedCallback;
import net.fabricmc.loader.api.FabricLoader;
import net.minecraft.class_1074;
import net.minecraft.class_1158;
import net.minecraft.class_1160;
import net.minecraft.class_1269;
import net.minecraft.class_1747;
import net.minecraft.class_1750;
import net.minecraft.class_1765;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_2248;
import net.minecraft.class_2323;
import net.minecraft.class_2378;
import net.minecraft.class_2533;
import net.minecraft.class_2561;
import net.minecraft.class_2588;
import net.minecraft.class_2960;
import net.minecraft.class_4587;
import net.minecraft.class_5250;
import net.minecraft.item.*;
import org.apache.commons.lang3.tuple.ImmutableTriple;
import org.apache.commons.lang3.tuple.Triple;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.util.function.QuadFunction;
import org.dimdev.dimdoors.block.door.DimensionalDoorBlock;
import org.dimdev.dimdoors.block.door.DimensionalTrapdoorBlock;
import org.dimdev.dimdoors.block.door.data.DoorData;
import org.dimdev.dimdoors.block.entity.EntranceRiftBlockEntity;
import org.dimdev.dimdoors.client.UnderlaidChildItemRenderer;
import org.dimdev.dimdoors.listener.ItemRegistryEntryAddedListener;
import org.dimdev.dimdoors.rift.targets.PublicPocketTarget;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.function.Consumer;
import java.util.function.Function;

public class DimensionalDoorItemRegistrar {
	public static final String PREFIX = "item_ag_dim_";

	private final class_2378<class_1792> registry;

	private final Map<class_2248, class_2248> blocksAlreadyNotifiedAbout = new HashMap<>();
	private final Map<class_2248, Triple<class_2960, class_1792, Function<class_2248, class_1747>>> toBeMapped = new HashMap<>();

	private final Map<class_1792, Function<class_1750, class_1269>> placementFunctions = new HashMap<>();

	public DimensionalDoorItemRegistrar(class_2378<class_1792> registry) {
		this.registry = registry;

		init();
		RegistryEntryAddedCallback.event(registry).register(new ItemRegistryEntryAddedListener(this));
	}

	public boolean isRegistered(class_1792 item) {
		return placementFunctions.containsKey(item);
	}

	public class_1269 place(class_1792 item, class_1750 context) {
		return placementFunctions.get(item).apply(context);
	}

	private void init() {
		new ArrayList<>(registry.method_29722())
				.forEach(entry -> handleEntry(entry.getKey().method_29177(), entry.getValue()));
	}

	public void handleEntry(class_2960 identifier, class_1792 original) {
		if (DimensionalDoorsInitializer.getConfig().getDoorsConfig().isAllowed(identifier)) {
			if (original instanceof class_1765) {
				class_2248 block = ((class_1765) original).method_7711();
				handleEntry(identifier, original, block, AutoGenDimensionalDoorItem::new);
			} else if (original instanceof class_1747) {
				class_2248 originalBlock = ((class_1747) original).method_7711();
				if (originalBlock instanceof class_2323) {
					handleEntry(identifier, original, originalBlock, AutoGenDimensionalDoorItem::new);
				} else {
					handleEntry(identifier, original, originalBlock, AutoGenDimensionalTrapdoorItem::new);
				}
			}
		}
	}

	private void handleEntry(class_2960 identifier, class_1792 original, class_2248 originalBlock, QuadFunction<class_2248, class_1792.class_1793, Consumer<? super EntranceRiftBlockEntity>, class_1792, ? extends class_1747> constructor) {

		if (!(originalBlock instanceof DimensionalDoorBlock)
				&& !(originalBlock instanceof DimensionalTrapdoorBlock)
				&& (originalBlock instanceof class_2323 || originalBlock instanceof class_2533)) {
			class_1792.class_1793 settings = ItemExtensions.getSettings(original).method_7892(DoorData.PARENT_ITEMS.contains(original) || DoorData.PARENT_BLOCKS.contains(originalBlock) ? null : ModItems.DIMENSIONAL_DOORS);

			Function<class_2248, class_1747> dimItemConstructor = (dimBlock) -> constructor.apply(dimBlock, settings, rift -> rift.setDestination(new PublicPocketTarget()), original);

			if (!blocksAlreadyNotifiedAbout.containsKey(originalBlock)) {
				toBeMapped.put(originalBlock, new ImmutableTriple<>(identifier, original, dimItemConstructor));
				return;
			}

			register(identifier, original, blocksAlreadyNotifiedAbout.get(originalBlock), dimItemConstructor);
		}
	}

	public void notifyBlockMapped(class_2248 original, class_2248 dimBlock) {
		if (!toBeMapped.containsKey(original)) {
			blocksAlreadyNotifiedAbout.put(original, dimBlock);
			return;
		}
		Triple<class_2960, class_1792, Function<class_2248, class_1747>> triple = toBeMapped.get(original);
		register(triple.getLeft(), triple.getMiddle(), dimBlock, triple.getRight());
	}

	private void register(class_2960 identifier, class_1792 original, class_2248 block, Function<class_2248, class_1747> dimItem) {
		if (!DoorData.PARENT_ITEMS.contains(original)) {
			class_2960 gennedId = new class_2960("dimdoors", PREFIX + identifier.method_12836() + "_" + identifier.method_12832());
			class_1747 item = class_2378.method_10230(registry, gennedId, dimItem.apply(block));
			placementFunctions.put(original, item::method_7712);
			if (FabricLoader.getInstance().getEnvironmentType() == EnvType.CLIENT) {
				registerItemRenderer(item);
			}
		}
	}

	@Environment(EnvType.CLIENT)
	private void registerItemRenderer(class_1747 dimItem) {
		BuiltinItemRendererRegistry.INSTANCE.register(dimItem, Renderer.RENDERER);
	}

	// extract renderer to inner interface so it can be removed in server environment via annotation
	@Environment(EnvType.CLIENT)
	private interface Renderer {
		UnderlaidChildItemRenderer RENDERER = new UnderlaidChildItemRenderer(class_1802.field_8634);
	}

	private static class AutoGenDimensionalDoorItem extends DimensionalDoorItem implements ChildItem {
		private final class_1792 originalItem;

		public AutoGenDimensionalDoorItem(class_2248 block, class_1793 settings, Consumer<? super EntranceRiftBlockEntity> setupFunction, class_1792 originalItem) {
			super(block, settings, setupFunction);
			this.originalItem = originalItem;
		}

		@Override
		public class_2561 method_7864(class_1799 stack) {
			return class_5250.method_43477(new class_2588("dimdoors.autogen_item_prefix", class_1074.method_4662(originalItem.method_7876())));
		}

		@Override
		public class_2561 method_7848() {
			return class_5250.method_43477(new class_2588("dimdoors.autogen_item_prefix", class_1074.method_4662(originalItem.method_7876())));
		}

		@Override
		public class_1792 getOriginalItem() {
			return originalItem;
		}

		@Override
		public void transform(class_4587 matrices) {
			matrices.method_22905(0.68f, 0.68f, 1);
			matrices.method_22904(0.05, 0.02, 0);
		}
	}

	private static class AutoGenDimensionalTrapdoorItem extends DimensionalTrapdoorItem implements ChildItem {
		private final class_1792 originalItem;

		public AutoGenDimensionalTrapdoorItem(class_2248 block, class_1793 settings, Consumer<? super EntranceRiftBlockEntity> setupFunction, class_1792 originalItem) {
			super(block, settings, setupFunction);
			this.originalItem = originalItem;
		}

		@Override
		public class_2561 method_7864(class_1799 stack) {
			return class_5250.method_43477(new class_2588("dimdoors.autogen_item_prefix", class_1074.method_4662(originalItem.method_7876())));
		}

		@Override
		public class_2561 method_7848() {
			return class_5250.method_43477(new class_2588("dimdoors.autogen_item_prefix", class_1074.method_4662(originalItem.method_7876())));
		}

		@Override
		public class_1792 getOriginalItem() {
			return originalItem;
		}

		@Override
		public void transform(class_4587 matrices) {
			matrices.method_22905(0.55f, 0.55f, 0.6f);
			matrices.method_22904(0.05, -0.05, 0.41);
			matrices.method_22907(class_1158.method_35823(new class_1160(90, 0, 0)));
		}
	}

	public interface ChildItem {
		class_1792 getOriginalItem();

		default void transform(class_4587 matrices) {
		}
	}
}
