package org.dimdev.dimdoors.network;

import net.fabricmc.fabric.api.networking.v1.PacketByteBufs;
import net.fabricmc.fabric.api.networking.v1.PacketSender;
import net.fabricmc.fabric.api.networking.v1.ServerPlayNetworking;
import net.minecraft.class_1268;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2540;
import net.minecraft.class_2960;
import net.minecraft.class_3222;
import net.minecraft.class_3244;
import net.minecraft.class_5321;
import net.minecraft.server.MinecraftServer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.api.item.ExtendedItem;
import org.dimdev.dimdoors.network.packet.c2s.HitBlockWithItemC2SPacket;
import org.dimdev.dimdoors.network.packet.c2s.NetworkHandlerInitializedC2SPacket;
import org.dimdev.dimdoors.network.packet.s2c.PlayerInventorySlotUpdateS2CPacket;
import org.dimdev.dimdoors.network.packet.s2c.SyncPocketAddonsS2CPacket;
import org.dimdev.dimdoors.world.ModDimensions;
import org.dimdev.dimdoors.world.level.registry.DimensionalRegistry;
import org.dimdev.dimdoors.world.pocket.PocketDirectory;
import org.dimdev.dimdoors.world.pocket.type.Pocket;
import org.dimdev.dimdoors.world.pocket.type.addon.AutoSyncedAddon;

import java.io.IOException;
import java.util.*;
import java.util.function.Supplier;

// each client has their own corresponding ServerPacketHandler, so feel free to add client specific data in here
public class ServerPacketHandler implements ServerPacketListener {
	private static final Logger LOGGER = LogManager.getLogger();

	private final class_3244 networkHandler;
	private final Set<class_2960> registeredChannels = new HashSet<>();
	private boolean initialized = false;

	private class_5321<class_1937> lastSyncedPocketWorld;
	private int lastSyncedPocketId = Integer.MIN_VALUE;
	private boolean pocketSyncDirty = true;

	public void init() {
		if (initialized) throw new RuntimeException("ServerPacketHandler has already been initialized.");
		initialized = true;
		registerReceiver(NetworkHandlerInitializedC2SPacket.ID, NetworkHandlerInitializedC2SPacket::new);
		registerReceiver(HitBlockWithItemC2SPacket.ID, HitBlockWithItemC2SPacket::new);
	}

	public static ServerPacketHandler get(class_3222 player) {
		return get(player.field_13987);
	}

	public static ServerPacketHandler get(class_3244 networkHandler) {
		return ((ExtendedServerPlayNetworkHandler) networkHandler).getDimDoorsPacketHandler();
	}

	public static boolean sendPacket(class_3222 player, SimplePacket<?> packet) {
		try {
			ServerPlayNetworking.send(player, packet.channelId(), packet.write(PacketByteBufs.create()));
			return true;
		} catch (IOException e) {
			LOGGER.error(e);
			return false;
		}
	}

	public boolean sendPacket(SimplePacket<?> packet) {
		return sendPacket(getPlayer(), packet);
	}

	public ServerPacketHandler(class_3244 networkHandler) {
		this.networkHandler = networkHandler;
	}

	private void registerReceiver(class_2960 channelName, Supplier<? extends SimplePacket<ServerPacketListener>> supplier) {
		ServerPlayNetworking.registerReceiver(networkHandler, channelName,
				(MinecraftServer server, class_3222 player, class_3244 handler, class_2540 buf, PacketSender responseSender) -> {
					try {
						supplier.get().read(buf).apply(this);
					} catch (IOException e) {
						LOGGER.error(e);
					}
				});
		registeredChannels.add(channelName);
	}

	private void unregisterReceiver(class_2960 channelName) {
		ServerPlayNetworking.unregisterReceiver(networkHandler, channelName);
		registeredChannels.remove(channelName);
	}

	public void unregister() {
		new HashSet<>(registeredChannels).forEach(this::unregisterReceiver);
	}

	public MinecraftServer getServer() {
		return ((ExtendedServerPlayNetworkHandler) networkHandler).dimdoorsGetServer();
	}

	public class_3222 getPlayer() {
		return networkHandler.field_14140;
	}

	// TODO: attach this to some event to detect other kinds teleportation
	public void syncPocketAddonsIfNeeded(class_1937 world, class_2338 pos) {
		if (!ModDimensions.isPocketDimension(world)) return;
		PocketDirectory directory = DimensionalRegistry.getPocketDirectory(world.method_27983());
		Pocket pocket = directory.getPocketAt(pos);
		if (pocket == null) return;
		if ((pocketSyncDirty || pocket.getId() != lastSyncedPocketId || !world.method_27983().method_29177().equals(lastSyncedPocketWorld.method_29177()))) {
			pocketSyncDirty = false;
			lastSyncedPocketId = pocket.getId();
			lastSyncedPocketWorld = world.method_27983();
			sendPacket(getPlayer(), new SyncPocketAddonsS2CPacket(world.method_27983(), directory.getGridSize(), pocket.getId(), pocket.getRange(), pocket.getAddonsInstanceOf(AutoSyncedAddon.class)));
		}
	}

	public void markPocketSyncDirty(int id) {
		if (lastSyncedPocketId == id) pocketSyncDirty = true;
	}

	public void sync(class_1799 stack, class_1268 hand) {
		if (hand == class_1268.field_5810) {
			sendPacket(new PlayerInventorySlotUpdateS2CPacket(45, stack));
		} else {
			sendPacket(new PlayerInventorySlotUpdateS2CPacket(getPlayer().method_31548().field_7545, stack));
		}
	}

	@Override
	public void onAttackBlock(HitBlockWithItemC2SPacket packet) {
		getServer().execute(() -> {
			class_1792 item = getPlayer().method_5998(packet.getHand()).method_7909();
			if (item instanceof ExtendedItem) {
				((ExtendedItem) item).onAttackBlock(getPlayer().field_6002, getPlayer(), packet.getHand(), packet.getPos(), packet.getDirection());
			}
		});
	}

	@Override
	public void onNetworkHandlerInitialized(NetworkHandlerInitializedC2SPacket packet) {
		syncPocketAddonsIfNeeded(getPlayer().field_6002, getPlayer().method_24515());
	}
}
