package org.dimdev.dimdoors.pockets.generator;

import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_1263;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2519;
import net.minecraft.class_2520;
import net.minecraft.class_2595;
import net.minecraft.class_2601;
import net.minecraft.class_2791;
import net.minecraft.class_3300;
import net.minecraft.class_3738;
import net.minecraft.server.MinecraftServer;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.pockets.TemplateUtils;
import org.dimdev.dimdoors.pockets.modifier.LazyModifier;
import org.dimdev.dimdoors.pockets.modifier.Modifier;
import org.dimdev.dimdoors.pockets.modifier.RiftManager;
import org.dimdev.dimdoors.world.pocket.type.LazyGenerationPocket;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

import java.util.*;
import java.util.stream.Collectors;

public abstract class LazyPocketGenerator extends PocketGenerator {
	private static final Logger LOGGER = LogManager.getLogger();

	public static boolean currentlyGenerating = false;
	public static Queue<class_2791> generationQueue = new LinkedList<>();


	protected List<LazyModifier> lazyModifierList = new ArrayList<>();

	public void generateChunk(LazyGenerationPocket pocket, class_2791 chunk) {
		lazyModifierList.forEach(modifier -> modifier.applyToChunk(pocket, chunk));
	}

	// LazyPocketGenerator handles attaching itself so that it can drop itself if it has already generated everything necessary.
	public void attachToPocket(LazyGenerationPocket pocket) {
		// We assume that this LazyPocketGenerator has not been cloned yet if the modifier list has any entries since it should be empty at this stage
		if (this.modifierList.size() > 0) throw new UnsupportedOperationException("Cannot attach LazyPocketGenerator that has not been cloned yet to pocket");
		pocket.attachGenerator(this);
	}

	@Override
	public PocketGenerator fromNbt(class_2487 nbt, class_3300 manager) {
		super.fromNbt(nbt, manager);

		if (nbt.method_10545("lazy_modifiers")) {
			class_2499 modifiersNbt = nbt.method_10554("lazy_modifiers", 10);
			for (int i = 0; i < modifiersNbt.size(); i++) {
				// TODO: skip deserialization of single Modifiers on Exception.
				// TODO: Modifier via ResourceManager
				lazyModifierList.add((LazyModifier) Modifier.deserialize(modifiersNbt.method_10602(i), manager));
			}
		}

		if (nbt.method_10545("lazy_modifier_references")) {
			class_2499 modifiersNbt = nbt.method_10554("lazy_modifier_references", NbtType.STRING);
			for (class_2520 nbtElement : modifiersNbt) {
				// TODO: skip deserialization of single Modifiers on Exception.
				// TODO: Modifier via ResourceManager
				lazyModifierList.add((LazyModifier) Modifier.deserialize(nbtElement, manager));
			}
		}

		return this;
	}

	@Override
	public class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {
		super.toNbtInternal(nbt, allowReference);

		if (lazyModifierList.size() > 0) {
			List<class_2520> lazyModNbts = lazyModifierList.stream().map(lazyModifier -> lazyModifier.toNbt(new class_2487(), allowReference)).collect(Collectors.toList());

			class_2499 lazyModifiersNbt = new class_2499();
			lazyModifiersNbt.addAll(lazyModNbts.stream().filter(class_2487.class::isInstance).collect(Collectors.toList()));
			nbt.method_10566("lazy_modifiers", lazyModifiersNbt);

			class_2499 lazyModifierReferences = new class_2499();
			lazyModifiersNbt.addAll(lazyModNbts.stream().filter(class_2519.class::isInstance).collect(Collectors.toList()));
			nbt.method_10566("lazy_modifier_references", lazyModifierReferences);
		}

		return nbt;
	}

	@Override
	public RiftManager getRiftManager(Pocket pocket) {
		if (pocket instanceof LazyGenerationPocket) {
			return new RiftManager(pocket, true);
		} else {
			return new RiftManager(pocket, false);
		}
	}

	public void attachLazyModifiers(Collection<LazyModifier> lazyModifiers) {
		this.lazyModifierList.addAll(lazyModifiers);
	}

	public LazyPocketGenerator cloneWithLazyModifiers(class_2338 originalOrigin) {
		LazyPocketGenerator clone = cloneWithEmptyModifiers(originalOrigin);
		clone.attachLazyModifiers(this.modifierList.stream().filter(LazyModifier.class::isInstance).map(LazyModifier.class::cast).collect(Collectors.toList()));
		return clone;
	}

	public LazyPocketGenerator cloneWithEmptyModifiers(class_2338 originalOrigin) {
		LazyPocketGenerator generator = getNewInstance();

		// Builder/ weight related stuff seems irrelevant here
		generator.setupLoot = this.setupLoot;

		return generator;
	}

	public void setSetupLoot(Boolean setupLoot) {
		this.setupLoot = setupLoot;
	}

	abstract public LazyPocketGenerator getNewInstance();

	public void setupChunk(Pocket pocket, class_2791 chunk, boolean setupLootTables) {
		MinecraftServer server = DimensionalDoorsInitializer.getServer();
		chunk.method_12021().stream().map(chunk::method_8321).forEach(blockEntity -> { // RiftBlockEntities should already be initialized here
			if (setupLootTables && blockEntity instanceof class_1263) {
				class_1263 inventory = (class_1263) blockEntity;
				server.method_18858(new class_3738(server.method_3780(), () -> {
					if (inventory.method_5442()) {
						if (blockEntity instanceof class_2595 || blockEntity instanceof class_2601) {
							TemplateUtils.setupLootTable(DimensionalDoorsInitializer.getWorld(pocket.getWorld()), blockEntity, inventory, LOGGER);
						}
					}
				}));
			}
		});
	}
}
