package org.dimdev.dimdoors.pockets.generator;

import com.google.common.collect.Multimap;
import com.mojang.serialization.Lifecycle;
import net.fabricmc.fabric.api.event.registry.FabricRegistryBuilder;
import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_1263;
import net.minecraft.class_2370;
import net.minecraft.class_2378;
import net.minecraft.class_2382;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2519;
import net.minecraft.class_2520;
import net.minecraft.class_2595;
import net.minecraft.class_2601;
import net.minecraft.class_2960;
import net.minecraft.class_3218;
import net.minecraft.class_3300;
import net.minecraft.class_5321;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.util.ReferenceSerializable;
import org.dimdev.dimdoors.api.util.ResourceUtil;
import org.dimdev.dimdoors.pockets.TemplateUtils;
import org.dimdev.dimdoors.pockets.modifier.Modifier;
import org.dimdev.dimdoors.pockets.modifier.RiftManager;
import org.dimdev.dimdoors.api.util.Location;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.api.util.Weighted;
import org.dimdev.dimdoors.api.util.math.Equation;
import org.dimdev.dimdoors.api.util.math.Equation.EquationParseException;
import org.dimdev.dimdoors.world.pocket.type.AbstractPocket;
import org.dimdev.dimdoors.world.pocket.type.LazyGenerationPocket;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

import java.util.*;
import java.util.function.Supplier;

public abstract class PocketGenerator implements Weighted<PocketGenerationContext>, ReferenceSerializable {
	private static final Logger LOGGER = LogManager.getLogger();
	public static final class_2378<PocketGeneratorType<? extends PocketGenerator>> REGISTRY = FabricRegistryBuilder.from(new class_2370<PocketGeneratorType<? extends PocketGenerator>>(class_5321.method_29180(new class_2960("dimdoors", "pocket_generator_type")), Lifecycle.stable(), null)).buildAndRegister();
	public static final String RESOURCE_STARTING_PATH = "pockets/generator"; //TODO: might want to restructure data packs

	private static final String defaultWeightEquation = "5"; // TODO: make config
	private static final int fallbackWeight = 5; // TODO: make config
	protected final List<Modifier> modifierList = new ArrayList<>();

	private String resourceKey = null;

	private class_2487 builderNbt;
	protected String weight = defaultWeightEquation;
	protected Equation weightEquation;
	protected Boolean setupLoot;

	private final List<String> tags = new ArrayList<>();

	public PocketGenerator() { }

	public PocketGenerator(String weight) {
		this.weight = weight;
		parseWeight();
	}

	public static PocketGenerator deserialize(class_2520 nbt, class_3300 manager) {
		switch (nbt.method_10711()) {
			case NbtType.COMPOUND: // It's a serialized Modifier
				return PocketGenerator.deserialize((class_2487) nbt, manager);
			case NbtType.STRING: // It's a reference to a resource location
				// TODO: throw if manager is null
				return ResourceUtil.loadReferencedResource(manager, RESOURCE_STARTING_PATH, nbt.method_10714(), ResourceUtil.NBT_READER.andThenComposable(nbtElement -> deserialize(nbtElement, manager)));
			default:
				throw new RuntimeException(String.format("Unexpected NbtType %d!", nbt.method_10711()));
		}
	}

	public static PocketGenerator deserialize(class_2520 nbt) {
		return deserialize(nbt, null);
	}

	public static PocketGenerator deserialize(class_2487 nbt, class_3300 manager) {
		class_2960 id = class_2960.method_12829(nbt.method_10558("type")); // TODO: return some NONE PocketGenerator if type cannot be found or deserialization fails.
		PocketGeneratorType<? extends PocketGenerator> type = REGISTRY.method_10223(id);
		if (type == null) {
			LOGGER.error("Could not deserialize PocketGenerator: " + nbt.toString());
			return null;
		}
		return type.fromNbt(nbt, manager);
	}

	public static PocketGenerator deserialize(class_2487 nbt) {
		return deserialize(nbt, null);
	}

	public static class_2520 serialize(PocketGenerator pocketGenerator, boolean allowReference) {
		return pocketGenerator.toNbt(new class_2487(), allowReference);
	}

	public static class_2520 serialize(PocketGenerator pocketGenerator) {
		return serialize(pocketGenerator, false);
	}

	private void parseWeight() {
		try {
			this.weightEquation = Equation.parse(weight);
		} catch (EquationParseException e) {
			LOGGER.error("Could not parse weight equation \"" + weight + "\", defaulting to default weight equation \"" + defaultWeightEquation + "\"", e);
			try {
				// FIXME: do we actually want to have it serialize to the broken String equation we input?
				this.weightEquation = Equation.newEquation(Equation.parse(defaultWeightEquation)::apply, stringBuilder -> stringBuilder.append(weight));
			} catch (EquationParseException equationParseException) {
				LOGGER.error("Could not parse default weight equation \"" + defaultWeightEquation + "\", defaulting to fallback weight \"" + fallbackWeight + "\"", equationParseException);
				// FIXME: do we actually want to have it serialize to the broken String equation we input?
				this.weightEquation = Equation.newEquation(stringDoubleMap -> (double) fallbackWeight, stringBuilder -> stringBuilder.append(weight));
			}
		}
	}

	public PocketGenerator fromNbt(class_2487 nbt, class_3300 manager) {
		if (nbt.method_10573("builder", NbtType.COMPOUND)) builderNbt = nbt.method_10562("builder");

		this.weight = nbt.method_10545("weight") ? nbt.method_10558("weight") : defaultWeightEquation;
		parseWeight();

		if (nbt.method_10545("setup_loot")) setupLoot = nbt.method_10577("setup_loot");

		if (nbt.method_10545("modifiers")) {
			class_2499 modifiersNbt = nbt.method_10554("modifiers", 10);
			for (int i = 0; i < modifiersNbt.size(); i++) {
				modifierList.add(Modifier.deserialize(modifiersNbt.method_10602(i), manager));
			}
		}

		if (nbt.method_10545("modifier_references")) {
			class_2499 modifiersNbt = nbt.method_10554("modifier_references", NbtType.STRING);
			for (class_2520 nbtElement : modifiersNbt) {
				modifierList.add(Modifier.deserialize(nbtElement, manager));
			}
		}

		if (nbt.method_10545("tags")) {
			class_2499 nbtList = nbt.method_10554("tags", NbtType.STRING);
			for (int i = 0; i < nbtList.size(); i++) {
				tags.add(nbtList.method_10608(i));
			}
		}
		return this;
	}

	public PocketGenerator fromNbt(class_2487 nbt) {
		return fromNbt(nbt, null);
	}

	public class_2520 toNbt(class_2487 nbt, boolean allowReference) {
		if (allowReference && this.resourceKey != null) {
			return class_2519.method_23256(this.resourceKey);
		}
		return toNbtInternal(nbt, allowReference);
	}

	protected class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {
		this.getType().toNbt(nbt);

		if (builderNbt != null) nbt.method_10566("builder", builderNbt);

		if (!weight.equals(defaultWeightEquation)) nbt.method_10582("weight", weight);

		if (setupLoot != null) nbt.method_10556("setup_loot", setupLoot);

		class_2499 modifiersNbt = new class_2499();
		class_2499 modifierReferences = new class_2499();
		for (Modifier modifier : modifierList) {
			class_2520 modNbt = modifier.toNbt(new class_2487(), allowReference);
			switch (modNbt.method_10711()) {
				case NbtType.COMPOUND:
					modifiersNbt.add(modNbt);
					break;
				case NbtType.STRING:
					modifierReferences.add(modNbt);
					break;
				default:
					throw new RuntimeException(String.format("Unexpected NbtType %d!", modNbt.method_10711()));
			}
		}
		if (modifiersNbt.size() > 0) nbt.method_10566("modifiers", modifiersNbt);
		if (modifierReferences.size() > 0) nbt.method_10566("modifier_references", modifierReferences);

		if (tags.size() > 0) {
			class_2499 nbtList = new class_2499();
			for (String nbtStr : tags) {
				nbtList.add(class_2519.method_23256(nbtStr));
			}
			nbt.method_10566("tags", nbtList);
		}

		return nbt;
	}

	public class_2520 toNbt(class_2487 nbt) {
		return toNbt(nbt, false);
	}

	public void processFlags(Multimap<String, String> flags) {
		// TODO: discuss some flag standardization
		Collection<String> reference = flags.get("reference");
		if (reference.stream().findFirst().map(string -> string.equals("local") || string.equals("global")).orElse(false)) {
			resourceKey = flags.get("resource_key").stream().findFirst().orElse(null);
		}
	}

	public abstract Pocket prepareAndPlacePocket(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder);

	public abstract PocketGeneratorType<? extends PocketGenerator> getType();

	public abstract String getKey();

	@Override
	public double getWeight(PocketGenerationContext parameters) {
		return this.weightEquation.apply(parameters.toVariableMap(new HashMap<>()));
	}

	public boolean isSetupLoot() {
		return setupLoot != null && setupLoot;
	}

	public void applyModifiers(PocketGenerationContext parameters, RiftManager manager) {
		for (Modifier modifier : modifierList) {
			modifier.apply(parameters, manager);
		}
	}

	public void applyModifiers(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder) {
		for (Modifier modifier : modifierList) {
			modifier.apply(parameters, builder);
		}
	}

	public void setup(Pocket pocket, RiftManager manager, PocketGenerationContext parameters, boolean setupLootTables) {
		class_3218 world = parameters.world();

		if (!(pocket instanceof LazyGenerationPocket)) { // should not iterate over that which does not exist & area may be massive, getBlockEntities() might force generation
			if (setupLootTables) // temp
				pocket.getBlockEntities().forEach((blockPos, blockEntity) -> {
					if (/*setupLootTables &&*/ blockEntity instanceof class_1263) { // comment in if needed
						class_1263 inventory = (class_1263) blockEntity;
						if (inventory.method_5442()) {
							if (blockEntity instanceof class_2595 || blockEntity instanceof class_2601) {
								TemplateUtils.setupLootTable(world, blockEntity, inventory, LOGGER);
								if (inventory.method_5442()) {
									LOGGER.error(", however Inventory is: empty!");
								}
							}
						}
					}
				});
		}
		manager.getRifts().forEach(rift -> rift.getDestination().setLocation(new Location(world, rift.method_11016())));
		TemplateUtils.registerRifts(manager.getRifts(), parameters.linkTo(), parameters.linkProperties(), pocket);
	}

	public RiftManager getRiftManager(Pocket pocket) {
		return new RiftManager(pocket);
	}

	// why would you want to check for exact tags, but still need a blackList? Good question, but there is probably some use case for it.
	public boolean checkTags(List<String> required, List<String> blackList, boolean exact) {
		if (exact && required.size() != tags.size()) return false;
		if (required != null) {
			for (String req : required) {
				if (!tags.contains(req)) return false;
			}
		}
		if (blackList != null) {
			for (String black : blackList) {
				if (tags.contains(black)) return false;
			}
		}
		return true;
	}

	public Pocket.PocketBuilder<?, ?> pocketBuilder(PocketGenerationContext parameters) { // TODO: PocketBuilder from json
		if (builderNbt == null){
			return Pocket.builder()
					.expand(getSize(parameters));
		}
		AbstractPocket.AbstractPocketBuilder<?, ?> abstractBuilder = AbstractPocket.deserializeBuilder(builderNbt);
		if (! (abstractBuilder instanceof Pocket.PocketBuilder)) {
			return Pocket.builder()
					.expand(getSize(parameters));
		}
		Pocket.PocketBuilder<?, ?> builder = (Pocket.PocketBuilder<?, ?>) abstractBuilder;
		return builder.expand(getSize(parameters));
	}

	public abstract class_2382 getSize(PocketGenerationContext parameters);

	public interface PocketGeneratorType<T extends PocketGenerator> {
		PocketGeneratorType<SchematicGenerator> SCHEMATIC = register(new class_2960("dimdoors", SchematicGenerator.KEY), SchematicGenerator::new);
		PocketGeneratorType<ChunkGenerator> CHUNK = register(new class_2960("dimdoors", ChunkGenerator.KEY), ChunkGenerator::new);
		PocketGeneratorType<VoidGenerator> VOID = register(new class_2960("dimdoors", VoidGenerator.KEY), VoidGenerator::new);

		PocketGenerator fromNbt(class_2487 nbt, class_3300 manager);

		class_2487 toNbt(class_2487 nbt);

		static void register() {
			DimensionalDoorsInitializer.apiSubscribers.forEach(d -> d.registerPocketGeneratorTypes(REGISTRY));
		}

		static <U extends PocketGenerator> PocketGeneratorType<U> register(class_2960 id, Supplier<U> constructor) {
			return class_2378.method_10230(REGISTRY, id, new PocketGeneratorType<U>() {
				@Override
				public PocketGenerator fromNbt(class_2487 nbt, class_3300 manager) {
					return constructor.get().fromNbt(nbt, manager);
				}

				@Override
				public class_2487 toNbt(class_2487 nbt) {
					nbt.method_10582("type", id.toString());
					return nbt;
				}
			});

		}
	}
}
