package org.dimdev.dimdoors.pockets.generator;

import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_2338;
import net.minecraft.class_2382;
import net.minecraft.class_2487;
import net.minecraft.class_2588;
import net.minecraft.class_2791;
import net.minecraft.class_2960;
import net.minecraft.class_3218;
import net.minecraft.class_3300;
import net.minecraft.class_5250;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.api.util.BlockPlacementType;
import org.dimdev.dimdoors.api.util.Path;
import org.dimdev.dimdoors.block.entity.RiftBlockEntity;
import org.dimdev.dimdoors.command.PocketCommand;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.pockets.PocketLoader;
import org.dimdev.dimdoors.pockets.PocketTemplate;
import org.dimdev.dimdoors.pockets.modifier.AbsoluteRiftBlockEntityModifier;
import org.dimdev.dimdoors.pockets.modifier.RiftManager;
import org.dimdev.dimdoors.util.schematic.Schematic;
import org.dimdev.dimdoors.world.level.registry.DimensionalRegistry;
import org.dimdev.dimdoors.world.pocket.type.LazyGenerationPocket;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class SchematicGenerator extends LazyPocketGenerator{
	private static final Logger LOGGER = LogManager.getLogger();
	public static final String KEY = "schematic";

	private String id;
	private class_2960 templateID;
	private BlockPlacementType placementType = BlockPlacementType.SECTION_NO_UPDATE;

	private final List<RiftBlockEntity> rifts = new ArrayList<>();
	private class_2338 origin;

	private AbsoluteRiftBlockEntityModifier queuedRiftBlockEntities;

	public SchematicGenerator() {
	}

	public SchematicGenerator(String id) {
		this.id = id;

		this.templateID = new class_2960("dimdoors", id);
	}

	public String getId() {
		return this.id;
	}

	public class_2960 getTemplateID() {
		return templateID;
	}

	@Override
	public void generateChunk(LazyGenerationPocket pocket, class_2791 chunk) {
		PocketTemplate template = PocketLoader.getInstance().getTemplates().get(Path.stringPath(templateID));
		if (template == null) throw new RuntimeException("Pocket template of id " + templateID + " not found!");
		template.place(pocket, chunk, origin, placementType);
		setupChunk(pocket, chunk, isSetupLoot());

		super.generateChunk(pocket, chunk);
	}

	@Override
	public PocketGenerator fromNbt(class_2487 nbt, class_3300 manager) {
		super.fromNbt(nbt, manager);

		this.id = nbt.method_10558("id"); // TODO: should we force having the "dimdoors:" in the json?
		this.templateID = new class_2960("dimdoors", id);
		if (nbt.method_10573("origin", NbtType.INT_ARRAY)) {
			int[] originInts = nbt.method_10561("origin");
			this.origin = new class_2338(originInts[0], originInts[1], originInts[2]);
		}
		if (nbt.method_10573("placement_type", NbtType.STRING)) placementType = BlockPlacementType.getFromId(nbt.method_10558("placement_type"));

		return this;
	}

	@Override
	public class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {
		super.toNbtInternal(nbt, allowReference);

		nbt.method_10582("id", this.id);
		if (placementType != BlockPlacementType.SECTION_NO_UPDATE) nbt.method_10582("placement_type", placementType.getId());

		if (origin != null) nbt.method_10539("origin", new int[]{origin.method_10263(), origin.method_10264(), origin.method_10260()});

		return nbt;
	}

	@Override
	public RiftManager getRiftManager(Pocket pocket) {
		RiftManager manager = super.getRiftManager(pocket);

		rifts.forEach(manager::add);

		return manager;
	}

	@Override
	public LazyPocketGenerator cloneWithLazyModifiers(class_2338 originalOrigin) {
		LazyPocketGenerator generator = super.cloneWithLazyModifiers(originalOrigin);
		generator.lazyModifierList.add(0, queuedRiftBlockEntities);

		return generator;
	}

	@Override
	public LazyPocketGenerator cloneWithEmptyModifiers(class_2338 originalOrigin) {
		SchematicGenerator generator = (SchematicGenerator) super.cloneWithEmptyModifiers(originalOrigin);

		generator.id = id;
		generator.templateID = templateID;
		generator.origin = originalOrigin;

		return generator;
	}

	@Override
	public LazyPocketGenerator getNewInstance() {
		return new SchematicGenerator();
	}

	@Override
	public Pocket prepareAndPlacePocket(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder) {
		class_3218 world = parameters.world();
		Map<String, Double> variableMap = parameters.toVariableMap(new HashMap<>());

		PocketTemplate template = PocketLoader.getInstance().getTemplates().get(Path.stringPath(templateID));
		if (template == null) throw new RuntimeException("Pocket template of id " + templateID + " not found!");

		Pocket pocket = DimensionalRegistry.getPocketDirectory(world.method_27983()).newPocket(builder);
		class_2338 origin = pocket.getOrigin();
		LOGGER.info("Generating pocket from template " + templateID + " at location " + origin);
		PocketCommand.logSetting.values().forEach(commandSource -> commandSource.method_9226(class_5250.method_43477(new class_2588("commands.pocket.log.creation.generating", templateID, origin.method_10263(), origin.method_10264(), origin.method_10260())), false));


		if (pocket instanceof LazyGenerationPocket) {
			Map<class_2338, RiftBlockEntity> absoluteRifts = template.getAbsoluteRifts(pocket);
			rifts.addAll(absoluteRifts.values());

			queuedRiftBlockEntities = new AbsoluteRiftBlockEntityModifier(absoluteRifts);
		} else {
			template.place(pocket, placementType);
		}

		return pocket;
	}

	@Override
	public PocketGeneratorType<? extends PocketGenerator> getType() {
		return PocketGeneratorType.SCHEMATIC;
	}

	@Override
	public String getKey() {
		return KEY;
	}

	@Override
	public class_2382 getSize(PocketGenerationContext parameters) {
		PocketTemplate template = PocketLoader.getInstance().getTemplates().get(Path.stringPath(templateID));
		if (template == null) throw new RuntimeException("Pocket template of id " + templateID + " not found!");
		Schematic schem = template.getSchematic();
		return new class_2382(schem.getWidth(), schem.getHeight(), schem.getLength());
	}
}
