package org.dimdev.dimdoors.pockets.generator;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.api.util.math.Equation;
import org.dimdev.dimdoors.api.util.math.Equation.EquationParseException;
import org.dimdev.dimdoors.world.level.registry.DimensionalRegistry;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

import java.util.HashMap;
import java.util.Map;
import net.minecraft.class_2338;
import net.minecraft.class_2382;
import net.minecraft.class_2487;
import net.minecraft.class_3300;

public class VoidGenerator extends LazyPocketGenerator {
	private static final Logger LOGGER = LogManager.getLogger();
	public static final String KEY = "void";
	private String width;
	private Equation heightEquation;
	private String height;
	private Equation widthEquation;
	private String length;
	private Equation lengthEquation;

	@Override
	public Pocket prepareAndPlacePocket(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder) {
		Pocket pocket = DimensionalRegistry.getPocketDirectory(parameters.world().method_27983()).newPocket(builder);
		Map<String, Double> variableMap = parameters.toVariableMap(new HashMap<>());
		pocket.setSize((int) widthEquation.apply(variableMap), (int) heightEquation.apply(variableMap), (int) lengthEquation.apply(variableMap));

		return pocket;
	}

	@Override
	public PocketGeneratorType<? extends PocketGenerator> getType() {
		return PocketGeneratorType.VOID;
	}

	@Override
	public String getKey() {
		return KEY;
	}

	@Override
	public class_2382 getSize(PocketGenerationContext parameters) {
		Map<String, Double> variableMap = parameters.toVariableMap(new HashMap<>());
		return new class_2382((int) widthEquation.apply(variableMap), (int) heightEquation.apply(variableMap), (int) lengthEquation.apply(variableMap));
	}

	@Override
	public PocketGenerator fromNbt(class_2487 nbt, class_3300 manager) {
		super.fromNbt(nbt, manager);

		try {
			width = nbt.method_10558("width");
			widthEquation = Equation.parse(width);
			height = nbt.method_10558("height");
			heightEquation = Equation.parse(height);
			length = nbt.method_10558("length");
			lengthEquation = Equation.parse(length);

		} catch (EquationParseException e) {
			LOGGER.error(e);
		}

		return this;
	}

	@Override
	public class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {
		super.toNbtInternal(nbt, allowReference);

		nbt.method_10582("width", width);
		nbt.method_10582("height", height);
		nbt.method_10582("length", length);

		return nbt;
	}

	@Override
	public LazyPocketGenerator cloneWithEmptyModifiers(class_2338 originalOrigin) {
		VoidGenerator generator = (VoidGenerator) super.cloneWithEmptyModifiers(originalOrigin);
		generator.width = width;
		generator.height = height;
		generator.length = length;

		return generator;
	}

	@Override
	public LazyPocketGenerator getNewInstance() {
		return new VoidGenerator();
	}
}
