package org.dimdev.dimdoors.pockets.modifier;

import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2586;
import net.minecraft.class_2791;
import net.minecraft.class_3300;
import net.minecraft.class_3341;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.util.BlockBoxUtil;
import org.dimdev.dimdoors.block.entity.RiftBlockEntity;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.world.pocket.type.LazyGenerationPocket;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

import java.util.Map;
import java.util.stream.Collectors;

public class AbsoluteRiftBlockEntityModifier extends AbstractLazyModifier {
	private static final Logger LOGGER = LogManager.getLogger();
	public static final String KEY = "block_entity";

	private Map<class_2338, RiftBlockEntity> rifts;
	private Map<class_2338, class_2487> serializedRifts;

	public AbsoluteRiftBlockEntityModifier() {
	}

	public AbsoluteRiftBlockEntityModifier(Map<class_2338, RiftBlockEntity> rifts) {
		this.rifts = rifts;

	}

	@Override
	public Modifier fromNbt(class_2487 nbt, class_3300 manager) {
		// TODO: rifts from resource
		serializedRifts = nbt.method_10554("rifts", NbtType.COMPOUND).parallelStream().unordered().map(class_2487.class::cast)
				.filter(compound -> {
					if (compound.method_10545("Pos")) {
						return true;
					}
					LOGGER.error("Discarding rift on deserialization since \"Pos\" tag was not set.");
					return false;
				})
				.collect(Collectors.toConcurrentMap(compound -> {
					int[] ints = compound.method_10561("Pos");
					return new class_2338(ints[0], ints[1], ints[2]);
				}, compound -> compound));

		return this;
	}

	@Override
	public class_2487 toNbtInternal(class_2487 nbt, boolean allowResource) {
		super.toNbtInternal(nbt, allowResource);

		class_2499 riftsNbt;
		if (rifts != null) {
			riftsNbt = rifts.values().parallelStream().unordered().map(rift -> {
				class_2487 e = new class_2487();
				rift.method_11007(e);
				return e;
			}).collect(Collectors.toCollection(class_2499::new));
		} else {
			riftsNbt = new class_2499();
			riftsNbt.addAll(serializedRifts.values());
		}
		nbt.method_10566("rifts", riftsNbt);

		return nbt;
	}

	@Override
	public ModifierType<? extends Modifier> getType() {
		return ModifierType.ABSOLUTE_RIFT_BLOCK_ENTITY_MODIFIER_TYPE;
	}

	@Override
	public String getKey() {
		return KEY;
	}

	@Override
	public void apply(PocketGenerationContext parameters, RiftManager manager) {
		if (!manager.isPocketLazy()) { // rifts is guaranteed to exist at this stage since this modifier is not supposed to be loaded from json
			class_1937 world = DimensionalDoorsInitializer.getWorld(manager.getPocket().getWorld());
			rifts.values().forEach(world::method_8438);
		}
	}

	@Override
	public void apply(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder) {

	}

	@Override
	public void applyToChunk(LazyGenerationPocket pocket, class_2791 chunk) {
		class_3341 chunkBox = BlockBoxUtil.getBox(chunk);

		if (rifts != null) {
			rifts.entrySet().stream().unordered().filter(entry -> chunkBox.method_14662(entry.getKey())).collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue))
					.forEach((pos, rift) -> {
						rifts.remove(pos);
						chunk.method_12007(rift);
					});
		} else {
			serializedRifts.entrySet().stream().unordered().filter(entry -> chunkBox.method_14662(entry.getKey())).collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue))
					.forEach((pos, riftNbt) -> {
						rifts.remove(pos);
						chunk.method_12007(class_2586.method_11005(pos, chunk.method_8320(pos), riftNbt));
					});
		}
	}
}
