package org.dimdev.dimdoors.pockets.modifier;

import java.util.HashMap;
import java.util.Map;

import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_1923;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2350;
import net.minecraft.class_2378;
import net.minecraft.class_2487;
import net.minecraft.class_2680;
import net.minecraft.class_2756;
import net.minecraft.class_2960;
import net.minecraft.class_3218;
import net.minecraft.class_3300;
import com.google.common.base.MoreObjects;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.block.door.DimensionalDoorBlock;
import org.dimdev.dimdoors.block.entity.EntranceRiftBlockEntity;
import org.dimdev.dimdoors.block.entity.ModBlockEntityTypes;
import org.dimdev.dimdoors.block.entity.RiftData;
import org.dimdev.dimdoors.pockets.PocketLoader;
import org.dimdev.dimdoors.rift.targets.IdMarker;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.api.util.NbtEquations;
import org.dimdev.dimdoors.api.util.math.Equation;
import org.dimdev.dimdoors.api.util.math.Equation.EquationParseException;
import org.dimdev.dimdoors.world.pocket.type.LazyGenerationPocket;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

public class DimensionalDoorModifier extends AbstractLazyCompatibleModifier {
	private static final Logger LOGGER = LogManager.getLogger();
	public static final String KEY = "door";

	private class_2350 facing;
	private String doorTypeString;
	private DimensionalDoorBlock doorType;
	private class_2487 doorData;
	private String doorDataReference;

	private String x;
	private String y;
	private String z;
	private Equation xEquation;
	private Equation yEquation;
	private Equation zEquation;

	@Override
	public Modifier fromNbt(class_2487 nbt, class_3300 manager) {
		String facingString = nbt.method_10558("facing");
		facing = class_2350.method_10168(nbt.method_10558("facing"));
		if (facing == null || facing.method_10166().method_10178()) {
			throw new RuntimeException("Could not interpret facing direction \"" + facingString + "\"");
		}

		doorTypeString = nbt.method_10558("door_type");
		class_2248 doorBlock = class_2378.field_11146.method_10223(class_2960.method_12829(doorTypeString));
		if (!(doorBlock instanceof DimensionalDoorBlock)) {
			throw new RuntimeException("Could not interpret door type \"" + doorTypeString + "\"");
		}
		doorType = (DimensionalDoorBlock) doorBlock;

		// TODO: rift data via ResourceManager
		if (nbt.method_10540("rift_data") == NbtType.STRING) {
			doorDataReference = nbt.method_10558("rift_data");
			doorData = PocketLoader.getInstance().getDataNbtCompound(doorDataReference);
		}

		else if (nbt.method_10540("rift_data") == NbtType.COMPOUND) doorData = nbt.method_10562("rift_data");

		try {
			x = nbt.method_10558("x");
			y = nbt.method_10558("y");
			z = nbt.method_10558("z");

			xEquation = Equation.parse(x);
			yEquation = Equation.parse(y);
			zEquation = Equation.parse(z);
		} catch (EquationParseException e) {
			LOGGER.error(e);
		}
		return this;
	}

	@Override
	public class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {
		super.toNbtInternal(nbt, allowReference);

		nbt.method_10582("facing", facing.method_15434());
		nbt.method_10582("door_type", doorTypeString);
		if (doorDataReference != null) nbt.method_10582("rift_data", doorDataReference);
		else if (doorData != null) nbt.method_10566("rift_data", doorData);
		nbt.method_10582("x", x);
		nbt.method_10582("y", y);
		nbt.method_10582("z", z);

		return nbt;
	}

	@Override
	public String toString() {
		return MoreObjects.toStringHelper(this)
				.add("facing", facing)
				.add("doorTypeString", doorTypeString)
				.add("doorType", doorType)
				.add("doorData", doorData)
				.add("doorDataReference", doorDataReference)
				.add("x", x)
				.add("y", y)
				.add("z", z)
				.add("xEquation", xEquation)
				.add("yEquation", yEquation)
				.add("zEquation", zEquation)
				.toString();
	}

	@Override
	public ModifierType<? extends Modifier> getType() {
		return ModifierType.DIMENSIONAL_DOOR_MODIFIER_TYPE;
	}

	@Override
	public String getKey() {
		return KEY;
	}

	@Override
	public void apply(PocketGenerationContext parameters, RiftManager manager) {
		Map<String, Double> variableMap = manager.getPocket().toVariableMap(new HashMap<>());
		class_2338 pocketOrigin = manager.getPocket().getOrigin();
		class_2338 pos = new class_2338(xEquation.apply(variableMap) + pocketOrigin.method_10263(), yEquation.apply(variableMap) + pocketOrigin.method_10264(), zEquation.apply(variableMap) + pocketOrigin.method_10260());

		class_2680 lower = doorType.method_9564().method_11657(DimensionalDoorBlock.field_10946, class_2756.field_12607).method_11657(DimensionalDoorBlock.field_10938, facing);
		class_2680 upper = doorType.method_9564().method_11657(DimensionalDoorBlock.field_10946, class_2756.field_12609).method_11657(DimensionalDoorBlock.field_10938, facing);
		EntranceRiftBlockEntity rift = ModBlockEntityTypes.ENTRANCE_RIFT.method_11032(pos, lower);

		if (doorData == null) {
			rift.setDestination(new IdMarker(manager.nextId()));
		} else {
			class_2487 solvedDoorData = NbtEquations.solveNbtCompoundEquations(doorData, variableMap);
			rift.setData(RiftData.fromNbt(solvedDoorData));
		}

		manager.add(rift);

		if (manager.getPocket() instanceof LazyGenerationPocket) {

			// queue two separate tasks, Cubic Chunks may cause the positions to be in different chunks.
			queueChunkModificationTask(new class_1923(pos), chunk -> {
				chunk.method_12010(pos, lower, false);
				chunk.method_12007(rift);
			});
			queueChunkModificationTask(new class_1923(pos.method_10084()), chunk -> {
				chunk.method_12010(pos.method_10084(), upper, false);
			});
		} else {
			class_3218 world = parameters.world();

			world.method_8501(pos, lower);
			world.method_8501(pos.method_10084(), upper);

			world.method_8438(rift);
		}
	}

	@Override
	public void apply(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder) {

	}
}
