package org.dimdev.dimdoors.pockets.modifier;

import com.google.common.collect.Multimap;
import com.mojang.serialization.Lifecycle;
import net.fabricmc.fabric.api.event.registry.FabricRegistryBuilder;
import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_2370;
import net.minecraft.class_2378;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_2960;
import net.minecraft.class_3300;
import net.minecraft.class_5321;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.util.ReferenceSerializable;
import org.dimdev.dimdoors.api.util.ResourceUtil;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

import java.util.Collection;
import java.util.function.Supplier;

public interface Modifier extends ReferenceSerializable {
	class_2378<ModifierType<? extends Modifier>> REGISTRY = FabricRegistryBuilder.from(new class_2370<ModifierType<? extends Modifier>>(class_5321.method_29180(new class_2960("dimdoors", "modifier_type")), Lifecycle.stable(), null)).buildAndRegister();

	String RESOURCE_STARTING_PATH = "pockets/modifier"; //TODO: might want to restructure data packs

	static Modifier deserialize(class_2520 nbt, class_3300 manager) {
		switch (nbt.method_10711()) {
			case NbtType.COMPOUND: // It's a serialized Modifier
				return Modifier.deserialize((class_2487) nbt, manager);
			case NbtType.STRING: // It's a reference to a resource location
				// TODO: throw if manager is null
				return ResourceUtil.loadReferencedResource(manager, RESOURCE_STARTING_PATH, nbt.method_10714(), ResourceUtil.NBT_READER.andThenComposable(nbtElement -> deserialize(nbtElement, manager)));
			default:
				throw new RuntimeException(String.format("Unexpected NbtType %d!", nbt.method_10711()));
		}
	}

	static Modifier deserialize(class_2520 nbt) {
		return deserialize(nbt, null);
	}

	static Modifier deserialize(class_2487 nbt, class_3300 manager) {
		class_2960 id = class_2960.method_12829(nbt.method_10558("type")); // TODO: return some NONE Modifier if type cannot be found or deserialization fails.
		return REGISTRY.method_10223(id).fromNbt(nbt, manager);
	}

	static Modifier deserialize(class_2487 nbt) {
		return deserialize(nbt, null);
	}

	static class_2520 serialize(Modifier modifier, boolean allowReference) {
		return modifier.toNbt(new class_2487(), allowReference);
	}

	static class_2520 serialize(Modifier modifier) {
		return serialize(modifier, false);
	}


	Modifier fromNbt(class_2487 nbt, class_3300 manager);

	default Modifier fromNbt(class_2487 nbt) {
		return fromNbt(nbt, null);
	}

	default class_2520 toNbt(class_2487 nbt, boolean allowReference) {
		return this.getType().toNbt(nbt);
	}

	default class_2520 toNbt(class_2487 nbt) {
		return toNbt(nbt, false);
	}

	void setResourceKey(String resourceKey);

	String getResourceKey();

	default void processFlags(Multimap<String, String> flags) {
		// TODO: discuss some flag standardization
		Collection<String> reference = flags.get("reference");
		if (reference.stream().findFirst().map(string -> string.equals("local") || string.equals("global")).orElse(false)) {
			setResourceKey(flags.get("resource_key").stream().findFirst().orElse(null));
		}
	}

	ModifierType<? extends Modifier> getType();

	String getKey();

	void apply(PocketGenerationContext parameters, RiftManager manager);

	void apply(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder);

	interface ModifierType<T extends Modifier> {
		ModifierType<ShellModifier> SHELL_MODIFIER_TYPE = register(new class_2960("dimdoors", ShellModifier.KEY), ShellModifier::new);
		ModifierType<DimensionalDoorModifier> DIMENSIONAL_DOOR_MODIFIER_TYPE = register(new class_2960("dimdoors", DimensionalDoorModifier.KEY), DimensionalDoorModifier::new);
		ModifierType<PocketEntranceModifier> PUBLIC_MODIFIER_TYPE = register(new class_2960("dimdoors", PocketEntranceModifier.KEY), PocketEntranceModifier::new);
		ModifierType<RiftDataModifier> RIFT_DATA_MODIFIER_TYPE = register(new class_2960("dimdoors", RiftDataModifier.KEY), RiftDataModifier::new);
		ModifierType<RelativeReferenceModifier> RELATIVE_REFERENCE_MODIFIER_TYPE = register(new class_2960("dimdoors", RelativeReferenceModifier.KEY), RelativeReferenceModifier::new);
		ModifierType<OffsetModifier> OFFSET_MODIFIER_TYPE = register(new class_2960("dimdoors", OffsetModifier.KEY), OffsetModifier::new);
		ModifierType<AbsoluteRiftBlockEntityModifier> ABSOLUTE_RIFT_BLOCK_ENTITY_MODIFIER_TYPE = register(new class_2960("dimdoors", AbsoluteRiftBlockEntityModifier.KEY), AbsoluteRiftBlockEntityModifier::new);

		Modifier fromNbt(class_2487 nbt, class_3300 manager);

		default Modifier fromNbt(class_2487 nbt) {
			return fromNbt(nbt, null);
		}

		class_2487 toNbt(class_2487 nbt);

		static void register() {
			DimensionalDoorsInitializer.apiSubscribers.forEach(d -> d.registerModifierTypes(REGISTRY));
		}

		static <U extends Modifier> ModifierType<U> register(class_2960 id, Supplier<U> factory) {
			return class_2378.method_10230(REGISTRY, id, new ModifierType<U>() {
				@Override
				public Modifier fromNbt(class_2487 nbt, class_3300 manager) {
					return factory.get().fromNbt(nbt, manager);
				}

				@Override
				public class_2487 toNbt(class_2487 nbt) {
					nbt.method_10582("type", id.toString());
					return nbt;
				}
			});
		}
	}
}
