package org.dimdev.dimdoors.pockets.modifier;

import java.util.Optional;
import net.minecraft.class_2487;
import net.minecraft.class_3218;
import net.minecraft.class_3300;
import com.google.common.base.MoreObjects;
import org.dimdev.dimdoors.block.entity.RiftBlockEntity;
import org.dimdev.dimdoors.rift.targets.LocalReference;
import org.dimdev.dimdoors.rift.targets.RiftReference;
import org.dimdev.dimdoors.api.util.Location;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

public class RelativeReferenceModifier extends AbstractModifier {
	public static final String KEY = "relative";

	private int point_a, point_b;

	@Override
	public Modifier fromNbt(class_2487 nbt, class_3300 manager) {
		point_a = nbt.method_10550("point_a");
		point_b = nbt.method_10550("point_b");
		return this;
	}

	@Override
	public class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {
		super.toNbtInternal(nbt, allowReference);
		nbt.method_10569("point_a", point_a);
		nbt.method_10569("point_b", point_b);
		return nbt;
	}

	@Override
	public ModifierType<? extends Modifier> getType() {
		return ModifierType.RELATIVE_REFERENCE_MODIFIER_TYPE;
	}

	@Override
	public String getKey() {
		return KEY;
	}

	@Override
	public void apply(PocketGenerationContext parameters, RiftManager manager) {
		Optional<Location> riftA = manager.get(point_a).map(rift -> new Location((class_3218) rift.method_10997(), rift.method_11016()));
		Optional<Location> riftB = manager.get(point_b).map(rift -> new Location((class_3218) rift.method_10997(), rift.method_11016()));

		if(riftA.isPresent() && riftB.isPresent()) {
			RiftReference link1 = LocalReference.tryMakeRelative(riftA.get(), riftB.get());
			RiftReference link2 = LocalReference.tryMakeRelative(riftB.get(), riftA.get());

			manager.consume(point_a, rift -> addLink(rift, link1));
			manager.consume(point_b, rift -> addLink(rift, link2));
		}
	}

	@Override
	public void apply(PocketGenerationContext parameters, Pocket.PocketBuilder<?, ?> builder) {

	}

	@Override
	public String toString() {
		return MoreObjects.toStringHelper(this)
				.add("point_a", point_a)
				.add("point_b", point_b)
				.toString();
	}

	private boolean addLink(RiftBlockEntity rift, RiftReference link) {
		rift.setDestination(link);
		return true;
	}
}
