package org.dimdev.dimdoors.pockets.virtual;

import com.mojang.serialization.*;
import net.fabricmc.fabric.api.event.registry.FabricRegistryBuilder;
import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_2370;
import net.minecraft.class_2378;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_2960;
import net.minecraft.class_3300;
import net.minecraft.class_5321;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.util.ResourceUtil;
import org.dimdev.dimdoors.pockets.virtual.reference.IdReference;
import org.dimdev.dimdoors.pockets.virtual.reference.PocketGeneratorReference;
import org.dimdev.dimdoors.pockets.virtual.reference.TagReference;
import org.dimdev.dimdoors.pockets.virtual.selection.ConditionalSelector;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.pockets.virtual.selection.PathSelector;
import org.dimdev.dimdoors.world.pocket.type.Pocket;
import org.jetbrains.annotations.Nullable;

import java.util.function.Supplier;

public interface ImplementedVirtualPocket extends VirtualPocket {
	String RESOURCE_STARTING_PATH = "pockets/virtual"; //TODO: might want to restructure data packs

	class_2378<VirtualPocketType<? extends ImplementedVirtualPocket>> REGISTRY = FabricRegistryBuilder.from(new class_2370<VirtualPocketType<? extends ImplementedVirtualPocket>>(class_5321.method_29180(new class_2960("dimdoors", "virtual_pocket_type")), Lifecycle.stable(), null)).buildAndRegister();

	static ImplementedVirtualPocket deserialize(class_2520 nbt, @Nullable class_3300 manager) {
		switch (nbt.method_10711()) {
			case NbtType.COMPOUND:
				return deserialize((class_2487) nbt, manager);
			case NbtType.STRING:
				return ResourceUtil.loadReferencedResource(manager, RESOURCE_STARTING_PATH, nbt.method_10714(), ResourceUtil.NBT_READER.andThenComposable(nbtElement -> deserialize(nbtElement, manager)));
			default:
				throw new RuntimeException(String.format("Unexpected NbtType %d!", nbt.method_10711()));
		}
	}

	static ImplementedVirtualPocket deserialize(class_2520 nbt) {
		return deserialize(nbt, null);
	}

	static ImplementedVirtualPocket deserialize(class_2487 nbt, @Nullable class_3300 manager) {
		class_2960 id = class_2960.method_12829(nbt.method_10558("type"));
		VirtualPocketType<?> type = REGISTRY.method_10223(id);
		return type != null ? type.fromNbt(nbt, manager) : VirtualPocketType.NONE.fromNbt(nbt, manager);
	}

	static ImplementedVirtualPocket deserialize(class_2487 nbt) {
		return deserialize(nbt, null);
	}

	static class_2520 serialize(ImplementedVirtualPocket implementedVirtualPocket, boolean allowReference) {
		return implementedVirtualPocket.toNbt(new class_2487(), allowReference);
	}

	static class_2520 serialize(ImplementedVirtualPocket implementedVirtualPocket) {
		return serialize(implementedVirtualPocket, false);
	}

	ImplementedVirtualPocket fromNbt(class_2487 nbt, @Nullable class_3300 manager);

	default ImplementedVirtualPocket fromNbt(class_2487 nbt) {
		return fromNbt(nbt, null);
	}

	class_2520 toNbt(class_2487 nbt, boolean allowReference);

	default class_2520 toNbt(class_2487 nbt) {
		return this.toNbt(nbt, false);
	}

	VirtualPocketType<? extends ImplementedVirtualPocket> getType();

	String getKey();

	interface VirtualPocketType<T extends ImplementedVirtualPocket> {
		VirtualPocketType<NoneVirtualPocket> NONE = register(new class_2960("dimdoors", NoneVirtualPocket.KEY), () -> NoneVirtualPocket.NONE);
		VirtualPocketType<IdReference> ID_REFERENCE = register(new class_2960("dimdoors", IdReference.KEY), IdReference::new);
		VirtualPocketType<TagReference> TAG_REFERENCE = register(new class_2960("dimdoors", TagReference.KEY), TagReference::new);
		VirtualPocketType<ConditionalSelector> CONDITIONAL_SELECTOR = register(new class_2960("dimdoors", ConditionalSelector.KEY), ConditionalSelector::new);
		VirtualPocketType<PathSelector> PATH_SELECTOR = register(new class_2960("dimdoors", PathSelector.KEY), PathSelector::new);

		ImplementedVirtualPocket fromNbt(class_2487 nbt, @Nullable class_3300 manager);

		default ImplementedVirtualPocket fromNbt(class_2487 nbt) {
			return fromNbt(nbt, null);
		}

		class_2487 toNbt(class_2487 nbt);

		static void register() {
			DimensionalDoorsInitializer.apiSubscribers.forEach(d -> d.registerVirtualSingularPocketTypes(REGISTRY));
		}

		static <U extends ImplementedVirtualPocket> VirtualPocketType<U> register(class_2960 id, Supplier<U> factory) {
			return class_2378.method_10230(REGISTRY, id, new VirtualPocketType<U>() {
				@Override
				public ImplementedVirtualPocket fromNbt(class_2487 nbt, class_3300 manager) {
					return factory.get().fromNbt(nbt, manager);
				}

				@Override
				public class_2487 toNbt(class_2487 nbt) {
					nbt.method_10582("type", id.toString());
					return nbt;
				}
			});
		}
	}

	// TODO: NoneReference instead?
	class NoneVirtualPocket extends AbstractVirtualPocket {
		public static final String KEY = "none";
		public static final NoneVirtualPocket NONE = new NoneVirtualPocket();

		private NoneVirtualPocket() {
		}

		@Override
		public Pocket prepareAndPlacePocket(PocketGenerationContext parameters) {
			throw new UnsupportedOperationException("Cannot place a NoneVirtualPocket");
		}

		@Override
		public PocketGeneratorReference getNextPocketGeneratorReference(PocketGenerationContext parameters) {
			throw new UnsupportedOperationException("Cannot get next pocket generator reference on a NoneVirtualPocket");
		}

		@Override
		public PocketGeneratorReference peekNextPocketGeneratorReference(PocketGenerationContext parameters) {
			throw new UnsupportedOperationException("Cannot peek next pocket generator reference on a NoneVirtualPocket");
		}

		@Override
		public ImplementedVirtualPocket fromNbt(class_2487 nbt, class_3300 manager) {
			return this;
		}

		@Override
		public VirtualPocketType<? extends ImplementedVirtualPocket> getType() {
			return VirtualPocketType.NONE;
		}

		@Override
		public String getKey() {
			return KEY;
		}

		@Override
		public double getWeight(PocketGenerationContext parameters) {
			return 0;
		}
	}
}
