package org.dimdev.dimdoors.pockets.virtual;

import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_2499;
import net.minecraft.class_2519;
import net.minecraft.class_2520;
import net.minecraft.class_3300;
import org.dimdev.dimdoors.api.util.ResourceUtil;
import org.dimdev.dimdoors.pockets.virtual.reference.PocketGeneratorReference;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.api.util.WeightedList;
import org.dimdev.dimdoors.world.pocket.type.Pocket;
import org.jetbrains.annotations.Nullable;

public class VirtualPocketList extends WeightedList<VirtualPocket, PocketGenerationContext> implements VirtualPocket {
	private String resourceKey = null;

	public static VirtualPocketList deserialize(class_2520 nbt, @Nullable class_3300 manager) {
		switch (nbt.method_10711()) {
			case NbtType.LIST:
				return deserialize((class_2499) nbt, manager);
			case NbtType.STRING:
				return ResourceUtil.loadReferencedResource(manager, RESOURCE_STARTING_PATH, nbt.method_10714(), ResourceUtil.NBT_READER.andThenComposable(nbtElement -> deserialize(nbtElement, manager)));
			default:
				throw new RuntimeException(String.format("Unexpected NbtType %d!", nbt.method_10711()));
		}
	}

	public static VirtualPocketList deserialize(class_2520 nbt) {
		return deserialize(nbt, null);
	}

	public static VirtualPocketList deserialize(class_2499 nbt, @Nullable class_3300 manager) {
		return new VirtualPocketList().fromNbt(nbt, manager);
	}

	public static VirtualPocketList deserialize(class_2499 nbt) {
		return deserialize(nbt, null);
	}



	@Override
	public void setResourceKey(String resourceKey) {
		this.resourceKey = resourceKey;
	}

	@Override
	public String getResourceKey() {
		return this.resourceKey;
	}

	public static class_2520 serialize(VirtualPocketList virtualPocketList, boolean allowReference) {
		return virtualPocketList.toNbt(new class_2499(), allowReference);
	}

	public static class_2520 serialize(VirtualPocketList virtualPocket) {
		return serialize(virtualPocket, false);
	}

	public VirtualPocketList() {
		super();
	}

	public VirtualPocketList fromNbt(class_2499 nbt, class_3300 manager) { // Keep in mind, this would add onto the list instead of overwriting it if called multiple times.
		for (class_2520 value : nbt) {
			this.add(VirtualPocket.deserialize(value, manager));
		}
		return this;
	}

	public VirtualPocketList fromNbt(class_2499 nbt) {
		return fromNbt(nbt, null);
	}

	public class_2520 toNbt(class_2499 nbt, boolean allowReference) {
		if (allowReference && resourceKey != null) {
			return class_2519.method_23256(resourceKey);
		}
		for(VirtualPocket virtualPocket : this) {
			nbt.add(VirtualPocket.serialize(virtualPocket, allowReference));
		}
		return nbt;
	}

	public class_2520 toNbt(class_2499 nbt) {
		return toNbt(nbt, false);
	}

	@Override
	public Pocket prepareAndPlacePocket(PocketGenerationContext context) {
		return getNextPocketGeneratorReference(context).prepareAndPlacePocket(context);
	}

	public PocketGeneratorReference getNextPocketGeneratorReference(PocketGenerationContext context) {
		return getNextRandomWeighted(context).getNextPocketGeneratorReference(context);
	}

	public PocketGeneratorReference peekNextPocketGeneratorReference(PocketGenerationContext context) {
		return peekNextRandomWeighted(context).peekNextPocketGeneratorReference(context);
	}

	@Override
	public void init() {
		this.forEach(VirtualPocket::init);
	}

	@Override
	public double getWeight(PocketGenerationContext context) {
		return getTotalWeight(context);
	}
}
