package org.dimdev.dimdoors.pockets.virtual.reference;

import net.fabricmc.fabric.api.util.NbtType;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_2519;
import net.minecraft.class_3300;
import com.google.common.base.MoreObjects;
import org.dimdev.dimdoors.pockets.PocketLoader;
import org.dimdev.dimdoors.pockets.generator.PocketGenerator;
import org.dimdev.dimdoors.pockets.virtual.ImplementedVirtualPocket;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.api.util.WeightedList;

import java.util.ArrayList;
import java.util.List;

public class TagReference extends PocketGeneratorReference {
	public static final String KEY = "tag";

	private final List<String> required = new ArrayList<>();
	private final List<String> blackList = new ArrayList<>();
	private Boolean exact;

	private WeightedList<PocketGenerator, PocketGenerationContext> pockets;

	@Override
	public ImplementedVirtualPocket fromNbt(class_2487 nbt, class_3300 manager) {
		super.fromNbt(nbt, manager);

		if (nbt.method_10545("required")) {
			class_2499 listNbt = nbt.method_10554("required", NbtType.STRING);
			for (int i = 0; i < listNbt.size(); i++) {
				required.add(listNbt.method_10608(i));
			}
		}

		if (nbt.method_10545("blackList")) {
			class_2499 listNbt = nbt.method_10554("blackList", NbtType.STRING);
			for (int i = 0; i < listNbt.size(); i++) {
				blackList.add(listNbt.method_10608(i));
			}
		}

		if (nbt.method_10545("exact")) exact = nbt.method_10577("exact");

		return this;
	}

	@Override
	protected class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {
		super.toNbtInternal(nbt, allowReference);

		if (required.size() > 0) {
			class_2499 listNbt = new class_2499();
			for (String nbtStr : required) {
				listNbt.add(class_2519.method_23256(nbtStr));
			}
			nbt.method_10566("required", listNbt);
		}

		if (blackList.size() > 0) {
			class_2499 list = new class_2499();
			for (String nbtStr : blackList) {
				list.add(class_2519.method_23256(nbtStr));
			}
			nbt.method_10566("blackList", list);
		}

		if (exact != null) {
			nbt.method_10556("exact", exact);
		}

		return nbt;
	}


	@Override
	public VirtualPocketType<? extends ImplementedVirtualPocket> getType() {
		return VirtualPocketType.TAG_REFERENCE;
	}

	@Override
	public String getKey() {
		return KEY;
	}
	// TODO: this will break if pockets change in between (which they could if we add a tool for creating pocket json config stuff ingame)
	@Override
	public PocketGenerator peekReferencedPocketGenerator(PocketGenerationContext parameters) {
		if (pockets == null) pockets = PocketLoader.getInstance().getPocketsMatchingTags(required, blackList, exact);
		return pockets.peekNextRandomWeighted(parameters);
	}

	@Override
	public PocketGenerator getReferencedPocketGenerator(PocketGenerationContext parameters) {
		if (pockets == null) pockets = PocketLoader.getInstance().getPocketsMatchingTags(required, blackList, exact);
		return pockets.getNextRandomWeighted(parameters);
	}

	@Override
	public String toString() {
		return MoreObjects.toStringHelper(this)
				.add("weight", weight)
				.add("weightEquation", weightEquation)
				.add("setupLoot", setupLoot)
				.add("modifierList", modifierList)
				.add("required", required)
				.add("blackList", blackList)
				.add("exact", exact)
				.add("pockets", pockets)
				.toString();
	}
}
