package org.dimdev.dimdoors.pockets.virtual.selection;

import com.google.common.collect.Maps;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.pockets.virtual.AbstractVirtualPocket;
import org.dimdev.dimdoors.pockets.virtual.ImplementedVirtualPocket;
import org.dimdev.dimdoors.pockets.virtual.VirtualPocket;
import org.dimdev.dimdoors.pockets.virtual.reference.PocketGeneratorReference;
import org.dimdev.dimdoors.pockets.PocketGenerationContext;
import org.dimdev.dimdoors.api.util.math.Equation;
import org.dimdev.dimdoors.world.pocket.type.Pocket;

import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import net.minecraft.class_2487;
import net.minecraft.class_2499;
import net.minecraft.class_3300;

public class ConditionalSelector extends AbstractVirtualPocket {
	private static final Logger LOGGER = LogManager.getLogger();
	public static final String KEY = "conditional";

	// TODO: redo this weird map part, Equations now have Equation.asString()
	private LinkedHashMap<String, VirtualPocket> pocketMap = Maps.newLinkedHashMap();
	private LinkedHashMap<String, Equation> equationMap = Maps.newLinkedHashMap();

	public ConditionalSelector() {
	}

	public ConditionalSelector(LinkedHashMap<String, VirtualPocket> pocketMap) {
		this.pocketMap = pocketMap;
	}

	public LinkedHashMap<String, VirtualPocket> getPocketMap() {
		return pocketMap;
	}

	@Override
	public ImplementedVirtualPocket fromNbt(class_2487 nbt, class_3300 manager) {
		class_2499 conditionalPockets = nbt.method_10554("pockets", 10);
		for (int i = 0; i < conditionalPockets.size(); i++) {
			class_2487 pocket = conditionalPockets.method_10602(i);
			String condition = pocket.method_10558("condition");
			if (pocketMap.containsKey(condition)) continue;
			try {
				equationMap.put(condition, Equation.parse(condition));
				pocketMap.put(condition, VirtualPocket.deserialize(pocket.method_10580("pocket"), manager));
			} catch (Equation.EquationParseException e) {
				LOGGER.error("Could not parse pocket condition equation!", e);
			}
		}
		return this;
	}

	@Override
	public class_2487 toNbtInternal(class_2487 nbt, boolean allowReference) {
		super.toNbtInternal(nbt, allowReference);

		class_2499 conditionalPockets = new class_2499();
		pocketMap.forEach((condition, pocket) -> {
			class_2487 compound = new class_2487();
			compound.method_10582("condition", condition);
			compound.method_10566("pocket", VirtualPocket.serialize(pocket, allowReference));
			conditionalPockets.add(compound);
		});
		nbt.method_10566("pockets", conditionalPockets);
		return nbt;
	}

	@Override
	public Pocket prepareAndPlacePocket(PocketGenerationContext parameters) {
		return getNextPocket(parameters).prepareAndPlacePocket(parameters);
	}

	@Override
	public PocketGeneratorReference getNextPocketGeneratorReference(PocketGenerationContext parameters) {
		return getNextPocket(parameters).getNextPocketGeneratorReference(parameters);
	}

	@Override
	public PocketGeneratorReference peekNextPocketGeneratorReference(PocketGenerationContext parameters) {
		return getNextPocket(parameters).peekNextPocketGeneratorReference(parameters);
	}

	@Override
	public void init() {
		pocketMap.values().forEach(VirtualPocket::init);
	}

	@Override
	public VirtualPocketType<? extends ImplementedVirtualPocket> getType() {
		return VirtualPocketType.CONDITIONAL_SELECTOR;
	}

	@Override
	public String getKey() {
		return KEY;
	}

	@Override
	public double getWeight(PocketGenerationContext parameters) {
		return getNextPocket(parameters).getWeight(parameters);
	}

	private VirtualPocket getNextPocket(PocketGenerationContext parameters) {
		for (Map.Entry<String, VirtualPocket> entry : pocketMap.entrySet()) {
			if (equationMap.get(entry.getKey()).asBoolean(parameters.toVariableMap(new HashMap<>()))) {
				return entry.getValue();
			}
		}
		return pocketMap.values().stream().findFirst().orElse(NoneVirtualPocket.NONE);
	}
}
