package org.dimdev.dimdoors.rift.registry;

import java.util.Objects;
import java.util.UUID;
import java.util.function.Function;
import net.fabricmc.fabric.api.event.registry.FabricRegistryBuilder;
import net.fabricmc.fabric.api.event.registry.RegistryAttribute;
import net.minecraft.class_1937;
import net.minecraft.class_2378;
import net.minecraft.class_2487;
import net.minecraft.class_2960;
import net.minecraft.class_5321;

public abstract class RegistryVertex {
	public static final class_2378<RegistryVertexType> registry = FabricRegistryBuilder.createSimple(RegistryVertex.RegistryVertexType.class, new class_2960("dimdoors", "registry_vertex")).attribute(RegistryAttribute.MODDED).buildAndRegister();

	private class_5321<class_1937> world; // The dimension to store this object in. Links are stored in both registries.

	protected UUID id = UUID.randomUUID(); // Used to create pointers to registry vertices. Should not be used for anything other than saving.

	public void sourceGone(RegistryVertex source) {
	}

	public void targetGone(RegistryVertex target) {
	}

	public void sourceAdded(RegistryVertex source) {
	}

	public void targetAdded(RegistryVertex target) {
	}

	public abstract RegistryVertexType<? extends RegistryVertex> getType();

	public String toString() {
		return "RegistryVertex(dim=" + this.world + ", id=" + this.id + ")";
	}

	public static RegistryVertex fromNbt(class_2487 nbt) {
		return Objects.requireNonNull(registry.method_10223(new class_2960(nbt.method_10558("type")))).fromNbt(nbt);
	}

	public static class_2487 toNbt(RegistryVertex registryVertex) {
		String type = registry.method_10221(registryVertex.getType()).toString();

		class_2487 nbt = registryVertex.getType().toNbt(registryVertex);
		nbt.method_10582("type", type);

		return nbt;
	}

	public UUID getId() {
		return id;
	}

	public void setId(UUID id) {
		this.id = id;
	}

	protected class_5321<class_1937> getWorld() {
		return world;
	}

	protected void setWorld(class_5321<class_1937> world) {
		this.world = world;
	}

	public interface RegistryVertexType<T extends RegistryVertex> {
		RegistryVertexType<PlayerRiftPointer> PLAYER = register("player", PlayerRiftPointer::fromNbt, PlayerRiftPointer::toNbt);
		RegistryVertexType<Rift> RIFT = register("rift", Rift::fromNbt, Rift::toNbt);
		RegistryVertexType<PocketEntrancePointer> ENTRANCE = register("entrance", PocketEntrancePointer::fromNbt, PocketEntrancePointer::toNbt);
		RegistryVertexType<RiftPlaceholder> RIFT_PLACEHOLDER = register("rift_placeholder", RiftPlaceholder::fromNbt, RiftPlaceholder::toNbt);

		T fromNbt(class_2487 nbt);

		class_2487 toNbt(RegistryVertex virtualType);

		static <T extends RegistryVertex> RegistryVertex.RegistryVertexType<T> register(String id, Function<class_2487, T> fromNbt, Function<T, class_2487> toNbt) {
			return class_2378.method_10226(registry, id, new RegistryVertexType<T>() {
				@Override
				public T fromNbt(class_2487 nbt) {
					return fromNbt.apply(nbt);
				}

				@Override
				public class_2487 toNbt(RegistryVertex registryVertex) {
					return toNbt.apply((T) registryVertex);
				}
			});
		}
	}
}
