package org.dimdev.dimdoors.rift.targets;

import com.mojang.serialization.Codec;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.rift.target.EntityTarget;
import org.dimdev.dimdoors.api.util.Location;
import org.dimdev.dimdoors.api.util.TeleportUtil;
import org.dimdev.dimdoors.block.ModBlocks;
import org.dimdev.dimdoors.block.UnravelUtil;
import org.dimdev.dimdoors.world.ModDimensions;

import java.util.UUID;
import net.minecraft.class_1297;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2379;
import net.minecraft.class_243;
import net.minecraft.class_2487;
import net.minecraft.class_2588;
import net.minecraft.class_3222;
import net.minecraft.class_5250;
import net.minecraft.class_5819;

import static org.dimdev.dimdoors.api.util.EntityUtils.chat;

public class EscapeTarget extends VirtualTarget implements EntityTarget { // TODO: createRift option
	private static final Logger LOGGER = LogManager.getLogger();

	public static final Codec<EscapeTarget> CODEC = RecordCodecBuilder.create(instance -> instance.group(
			Codec.BOOL.fieldOf("canEscapeLimbo").forGetter(target -> target.canEscapeLimbo)
	).apply(instance, EscapeTarget::new));

	protected final boolean canEscapeLimbo;

	public EscapeTarget(boolean canEscapeLimbo) {
		this.canEscapeLimbo = canEscapeLimbo;
	}

	@Override
	public boolean receiveEntity(class_1297 entity, class_243 relativePos, class_2379 relativeAngle, class_243 relativeVelocity) {
		if (!ModDimensions.isPocketDimension(entity.field_6002) && !(ModDimensions.isLimboDimension(entity.field_6002))) {
			chat(entity, class_5250.method_43477(new class_2588("rifts.destinations.escape.not_in_pocket_dim")));
			return false;
		}
		if (ModDimensions.isLimboDimension(entity.field_6002) && !this.canEscapeLimbo) {
			chat(entity, class_5250.method_43477(new class_2588("rifts.destinations.escape.cannot_escape_limbo")));
			return false;
		}
		if (entity.method_5770().field_9236)
			return false;
		UUID uuid = entity.method_5667();
		if (uuid != null) {
			//Location destLoc = DimensionalRegistry.getRiftRegistry().getOverworldRift(uuid);
			if (entity.field_6002.method_18470(uuid) == null) {
				LOGGER.log(Level.ERROR, "Tried to get player for escape target from uuid, but player does not exist, uh oh");
				return false;
			}
			LOGGER.log(Level.INFO, "sending player from limbo to their spawnpoint, good luck!");
			Location destLoc;
			if (((class_3222) entity.field_6002.method_18470(uuid)).method_26280() != null) {
				destLoc = new Location(((class_3222) entity.field_6002.method_18470(uuid)).method_26281(), ((class_3222) entity.field_6002.method_18470(uuid)).method_26280());
			} else {
				destLoc = new Location(DimensionalDoorsInitializer.getServer().method_30002(), DimensionalDoorsInitializer.getServer().method_30002().method_43126());

			}


			/*
			if (destLoc != null && destLoc.getBlockEntity() instanceof RiftBlockEntity || this.canEscapeLimbo) {
				//Location location = VirtualLocation.fromLocation(new Location((ServerWorld) entity.world, entity.getBlockPos())).projectToWorld(false);
				TeleportUtil.teleport(entity, destLoc.getWorld(), destLoc.getBlockPos(), relativeAngle, relativeVelocity);
			} else {
				if (destLoc == null) {
					chat(entity, MutableText.of(new TranslatableTextContent("rifts.destinations.escape.did_not_use_rift"));
				} else {
					chat(entity, MutableText.of(new TranslatableTextContent("rifts.destinations.escape.rift_has_closed"));
				}
				if (ModDimensions.LIMBO_DIMENSION != null) {
					TeleportUtil.teleport(entity, ModDimensions.LIMBO_DIMENSION, new BlockPos(this.location.getX(), this.location.getY(), this.location.getZ()), relativeAngle, relativeVelocity);
				}
			}
			 */


			if (destLoc != null && this.canEscapeLimbo) {
				Location location = destLoc; //VirtualLocation.fromLocation(new Location((ServerWorld) entity.world, destLoc.pos)).projectToWorld(false); //TODO Fix world projection.
				entity = TeleportUtil.teleport(entity, location.getWorld(), location.getBlockPos(), relativeAngle, relativeVelocity);
				entity.field_6017 = 0;
				class_5819 random = class_5819.method_43047();
				class_2338.method_25996(location.pos.method_10069(0, -3, 0), 3, 2, 3).forEach((pos1 -> {
					if (random.method_43057() < (1 / ((float) location.pos.method_10262(pos1))) * DimensionalDoorsInitializer.getConfig().getLimboConfig().limboBlocksCorruptingOverworldAmount) {
						class_2248 block = location.getWorld().method_8320(pos1).method_26204();
						if (UnravelUtil.unravelBlocksMap.containsKey(block))
							location.getWorld().method_8501(pos1, UnravelUtil.unravelBlocksMap.get(block).method_9564());
						else if (UnravelUtil.whitelistedBlocksForLimboRemoval.contains(block)) {
							location.getWorld().method_8501(pos1, ModBlocks.UNRAVELLED_FABRIC.method_9564());
						}
					}
				}));
			} else {
				if (destLoc == null) {
					chat(entity, class_5250.method_43477(new class_2588("rifts.destinations.escape.did_not_use_rift")));
				} else {
					chat(entity, class_5250.method_43477(new class_2588("rifts.destinations.escape.rift_has_closed")));
				}
				if (ModDimensions.LIMBO_DIMENSION != null) {
					entity = TeleportUtil.teleport(entity, ModDimensions.LIMBO_DIMENSION, new class_2338(this.location.getX(), this.location.getY(), this.location.getZ()), relativeAngle, relativeVelocity);
					entity.field_6017 = 0;
				}
			}
			return true;


		} else {
			return false; // No escape info for that entity
		}
	}

	@Override
	public VirtualTargetType<? extends VirtualTarget> getType() {
		return VirtualTargetType.ESCAPE;
	}

	public static class_2487 toNbt(EscapeTarget virtualTarget) {
		class_2487 nbt = new class_2487();
		nbt.method_10556("canEscapeLimbo", virtualTarget.canEscapeLimbo);
		return nbt;
	}

	public static EscapeTarget fromNbt(class_2487 nbt) {
		return new EscapeTarget(nbt.method_10577("canEscapeLimbo"));
	}
}
