package org.dimdev.dimdoors.rift.targets;

import java.util.Map;
import java.util.Objects;
import java.util.function.Function;

import com.mojang.serialization.Lifecycle;
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.rift.target.Target;
import org.dimdev.dimdoors.api.util.Location;
import org.dimdev.dimdoors.api.util.RGBA;
import net.fabricmc.fabric.api.event.registry.FabricRegistryBuilder;
import net.minecraft.class_2370;
import net.minecraft.class_2378;
import net.minecraft.class_2487;
import net.minecraft.class_2960;
import net.minecraft.class_5321;

/**
 * A target that is not an actual object in the game such as a block or a block
 * entity. Only virtual targets can be saved to NBT.
 */
public abstract class VirtualTarget implements Target {
	public static final class_2378<VirtualTargetType<?>> REGISTRY = FabricRegistryBuilder.<VirtualTargetType<?>, class_2370<VirtualTargetType<?>>>from(new class_2370<>(class_5321.method_29180(new class_2960("dimdoors", "virtual_type")), Lifecycle.stable(), null)).buildAndRegister();
	public static final RGBA COLOR = new RGBA(1, 0, 0, 1);

	protected Location location;

	public static VirtualTarget fromNbt(class_2487 nbt) {
		class_2960 id = new class_2960(nbt.method_10558("type"));
		return Objects.requireNonNull(REGISTRY.method_10223(id), "Unknown virtual target type " + id).fromNbt(nbt);
	}

	public static class_2487 toNbt(VirtualTarget virtualTarget) {
		class_2960 id = REGISTRY.method_10221(virtualTarget.getType());
		String type = id.toString();

		class_2487 nbt = virtualTarget.getType().toNbt(virtualTarget);
		nbt.method_10582("type", type);

		return nbt;
	}

	public void register() {
	}

	public void unregister() {
	}

	public abstract VirtualTargetType<? extends VirtualTarget> getType();

	public boolean shouldInvalidate(Location riftDeleted) {
		return false;
	}

	public RGBA getColor() {
		return this.getType().getColor();
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) return true;
		if (o == null || this.getClass() != o.getClass()) return false;
		VirtualTarget that = (VirtualTarget) o;
		return Objects.equals(this.location, that.location);
	}

	@Override
	public int hashCode() {
		return Objects.hash(this.location);
	}

	public void setLocation(Location location) {
		this.location = location;
	}

	public Location getLocation() {
		return this.location;
	}

	public boolean isDummy() {
		return false;
	}

	public interface VirtualTargetType<T extends VirtualTarget> {
		VirtualTargetType<RandomTarget> AVAILABLE_LINK = register("dimdoors:available_link", RandomTarget::fromNbt, RandomTarget::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<RandomTarget> DUNGEON = register("dimdoors:dungeon", DungeonTarget::fromNbt, DungeonTarget::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<EscapeTarget> ESCAPE = register("dimdoors:escape", EscapeTarget::fromNbt, EscapeTarget::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<GlobalReference> GLOBAL = register("dimdoors:global", GlobalReference::fromNbt, GlobalReference::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<LimboTarget> LIMBO = register("dimdoors:limbo", a -> LimboTarget.INSTANCE, a -> new class_2487(), VirtualTarget.COLOR);
		VirtualTargetType<LocalReference> LOCAL = register("dimdoors:local", LocalReference::fromNbt, LocalReference::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<PublicPocketTarget> PUBLIC_POCKET = register("dimdoors:public_pocket", PublicPocketTarget::fromNbt, PublicPocketTarget::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<PocketEntranceMarker> POCKET_ENTRANCE = register("dimdoors:pocket_entrance", PocketEntranceMarker::fromNbt, PocketEntranceMarker::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<PocketExitMarker> POCKET_EXIT = register("dimdoors:pocket_exit", a -> new PocketExitMarker(), a -> new class_2487(), VirtualTarget.COLOR);
		VirtualTargetType<PrivatePocketTarget> PRIVATE = register("dimdoors:private", a -> new PrivatePocketTarget(), a -> new class_2487(), PrivatePocketExitTarget.COLOR);
		VirtualTargetType<PrivatePocketExitTarget> PRIVATE_POCKET_EXIT = register("dimdoors:private_pocket_exit", a -> new PrivatePocketExitTarget(), a -> new class_2487(), PrivatePocketExitTarget.COLOR);
		VirtualTargetType<RelativeReference> RELATIVE = register("dimdoors:relative", RelativeReference::fromNbt, RelativeReference::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<IdMarker> ID_MARKER = register("dimdoors:id_marker", IdMarker::fromNbt, IdMarker::toNbt, VirtualTarget.COLOR);
		VirtualTargetType<UnstableTarget> UNSTABLE = register("dimdoors:unstable", nbt -> new UnstableTarget(), t -> new class_2487(), VirtualTarget.COLOR);
		VirtualTargetType<NoneTarget> NONE = register("dimdoors:none", nbt -> NoneTarget.INSTANCE, i -> new class_2487(), COLOR);
		Map<VirtualTargetType<?>, String> TRANSLATION_KEYS = new Object2ObjectArrayMap<>();

		T fromNbt(class_2487 nbt);

		class_2487 toNbt(VirtualTarget virtualType);

		RGBA getColor();

		default class_2960 getId() {
			return REGISTRY.method_10221(this);
		}

		default String getTranslationKey() {
			return TRANSLATION_KEYS.computeIfAbsent(this, t -> {
				class_2960 id = t.getId();
				return "dimdoors.virtualTarget." + id.method_12836() + "." + id.method_12832();
			});
		}

		static void register() {
			DimensionalDoorsInitializer.apiSubscribers.forEach(d -> d.registerVirtualTargetTypes(REGISTRY));
		}

		@SuppressWarnings("unchecked")
		static <T extends VirtualTarget> VirtualTargetType<T> register(String id, Function<class_2487, T> fromNbt, Function<T, class_2487> toNbt, RGBA color) {
			return class_2378.method_10226(REGISTRY, (String) id, new VirtualTargetType<T>() {
				@Override
				public T fromNbt(class_2487 nbt) {
					return fromNbt.apply(nbt);
				}

				@Override
				public class_2487 toNbt(VirtualTarget virtualType) {
					return toNbt.apply((T) virtualType);
				}

				@Override
				public RGBA getColor() {
					return color;
				}
			});
		}
	}

	@SuppressWarnings("EqualsWhichDoesntCheckParameterClass")
	public static class NoneTarget extends VirtualTarget {
		public static final NoneTarget INSTANCE = new NoneTarget();

		private NoneTarget() {
		}

		@Override
		public VirtualTargetType<? extends VirtualTarget> getType() {
			return VirtualTargetType.NONE;
		}

		@Override
		public boolean equals(Object o) {
			return o == INSTANCE;
		}

		@Override
		public int hashCode() {
			return System.identityHashCode(INSTANCE);
		}

		@Override
		public String toString() {
			return "[none]";
		}
	}
}
