package org.dimdev.dimdoors.world.decay;

import net.fabricmc.fabric.api.event.Event;
import net.fabricmc.fabric.api.event.EventFactory;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2680;
import java.util.Set;

public class DecayPattern {
    public static final Event<EntropyEvent> ENTROPY_EVENT = EventFactory.createArrayBacked(EntropyEvent.class, (world, pos, entorpy) -> {}, entropyEvents -> (world, pos, entorpy) -> {
        for (EntropyEvent event : entropyEvents) event.entropy(world, pos, entorpy);
    });

    private DecayPredicate predicate;
    private DecayProcessor processor;

    public DecayPattern() {}

    public DecayPattern(DecayPredicate predicate, DecayProcessor processor) {
        this.predicate = predicate;
        this.processor = processor;
    }

    public static DecayPattern deserialize(class_2487 nbt) {
        DecayPredicate predicate = DecayPredicate.deserialize(nbt.method_10562("predicate"));
        DecayProcessor processor = DecayProcessor.deserialize(nbt.method_10562("processor"));
        return DecayPattern.builder().predicate(predicate).processor(processor).create();
    }

    public boolean test(class_1937 world, class_2338 pos, class_2680 origin, class_2680 target) {
        return predicate.test(world, pos, origin, target);
    }

    public void process(class_1937 world, class_2338 pos, class_2680 origin, class_2680 target) {
        ENTROPY_EVENT.invoker().entropy(world, pos, processor.process(world, pos, origin, target));
    }

	public Set<class_2248> constructApplicableBlocks() {
		return predicate.constructApplicableBlocks();
	}

    public static DecayPattern.Builder builder() {
        return new DecayPattern.Builder();
    }

    public static class Builder {
        private DecayPredicate predicate = DecayPredicate.NONE;
        private DecayProcessor processor = DecayProcessor.NONE;

        public Builder predicate(DecayPredicate predicate) {
            this.predicate = predicate;
            return this;
        }

        public Builder processor(DecayProcessor processor) {
            this.processor = processor;
            return this;
        }

        public DecayPattern create() {
            return new DecayPattern(predicate, processor);
        }
    }

    private interface EntropyEvent {
        void entropy(class_1937 world, class_2338 pos, int entorpy);
    }
}
