package org.dimdev.dimdoors.world.pocket;

import java.util.Map;
import java.util.UUID;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;
import net.minecraft.class_1937;
import net.minecraft.class_2378;
import net.minecraft.class_2487;
import net.minecraft.class_2520;
import net.minecraft.class_2960;
import net.minecraft.class_3545;
import net.minecraft.class_5321;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import org.dimdev.dimdoors.world.level.registry.DimensionalRegistry;
import org.dimdev.dimdoors.world.pocket.type.Pocket;
import org.dimdev.dimdoors.world.pocket.type.PrivatePocket;

public class PrivateRegistry {
	protected static class PocketInfo {
		public final class_5321<class_1937> world;
		public final int id;

		public PocketInfo(class_5321<class_1937> world, int id) {
			this.world = world;
			this.id = id;
		}

		public static class_2487 toNbt(PocketInfo info) {
			class_2487 nbt = new class_2487();
			nbt.method_10582("world", info.world.method_29177().toString());
			nbt.method_10569("id", info.id);
			return nbt;
		}

		public static PocketInfo fromNbt(class_2487 nbt) {
			return new PocketInfo(
					class_5321.method_29179(class_2378.field_25298, new class_2960(nbt.method_10558("world"))),
					nbt.method_10550("id")
			);
		}
	}

	private static final String DATA_NAME = "dimdoors_private_pockets";

	protected BiMap<UUID, PocketInfo> privatePocketMap = HashBiMap.create(); // Player UUID -> Pocket Info TODO: fix AnnotatedNBT and use UUID rather than String

	public PrivateRegistry() {
	}

	public void fromNbt(class_2487 nbt) {
		privatePocketMap.clear();
		class_2487 privatePocketMapNbt = nbt.method_10562("private_pocket_map");
		CompletableFuture<Map<UUID, PocketInfo>> futurePrivatePocketMap = CompletableFuture.supplyAsync(() ->
				privatePocketMapNbt.method_10541().stream().unordered().map(key -> {
					class_2487 pocketInfoNbt = privatePocketMapNbt.method_10562(key);
					return CompletableFuture.supplyAsync(() -> new class_3545<>(UUID.fromString(key), PocketInfo.fromNbt(pocketInfoNbt)));
				}).parallel().map(CompletableFuture::join).collect(Collectors.toConcurrentMap(class_3545::method_15442, class_3545::method_15441)));

		futurePrivatePocketMap.join().forEach(this.privatePocketMap::put);
	}

	public class_2487 toNbt(class_2487 nbt) {
		CompletableFuture<class_2487> futurePrivatePocketMapNbt = CompletableFuture.supplyAsync(() -> {
			Map<String, class_2520> privatePocketNbtMap = this.privatePocketMap.entrySet().parallelStream().unordered().collect(Collectors.toConcurrentMap(entry -> entry.getKey().toString(), entry -> PocketInfo.toNbt(entry.getValue())));
			class_2487 privatePocketMapNbt = new class_2487();
			privatePocketNbtMap.forEach(privatePocketMapNbt::method_10566);
			return privatePocketMapNbt;
		});

		nbt.method_10566("private_pocket_map", futurePrivatePocketMapNbt.join());

		return nbt;
	}

	public PrivatePocket getPrivatePocket(UUID playerUUID) {
		PocketInfo pocket = this.privatePocketMap.get(playerUUID);
		if (pocket == null) return null;
		return DimensionalRegistry.getPocketDirectory(pocket.world).getPocket(pocket.id, PrivatePocket.class);
	}

	public void setPrivatePocketID(UUID playerUUID, Pocket pocket) {
		this.privatePocketMap.put(playerUUID, new PocketInfo(pocket.getWorld(), pocket.getId()));
	}

	public UUID getPrivatePocketOwner(Pocket pocket) {
		return this.privatePocketMap.inverse().get(new PocketInfo(pocket.getWorld(), pocket.getId()));
	}
}
