package org.dimdev.dimdoors.world.pocket.type;

import com.mojang.serialization.Lifecycle;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;

import net.fabricmc.fabric.api.event.registry.FabricRegistryBuilder;
import net.minecraft.class_1937;
import net.minecraft.class_2370;
import net.minecraft.class_2378;
import net.minecraft.class_2382;
import net.minecraft.class_2487;
import net.minecraft.class_2960;
import net.minecraft.class_5321;
import org.dimdev.dimdoors.world.pocket.PocketDirectory;

import java.util.Map;
import java.util.function.Supplier;

public abstract class AbstractPocket<V extends AbstractPocket<?>> {
	public static final class_2378<AbstractPocketType<? extends AbstractPocket<?>>> REGISTRY = FabricRegistryBuilder.from(new class_2370<AbstractPocketType<? extends AbstractPocket<?>>>(class_5321.method_29180(new class_2960("dimdoors", "abstract_pocket_type")), Lifecycle.stable(), null)).buildAndRegister();

	protected Integer id;
	protected class_5321<class_1937> world;

	public AbstractPocket(int id, class_5321<class_1937> world) {
		this.id = id;
		this.world = world;
	}

	protected AbstractPocket() {
	}

	public int getId() {
		return id;
	}

	public static AbstractPocket<? extends AbstractPocket<?>> deserialize(class_2487 nbt) {
		class_2960 id = class_2960.method_12829(nbt.method_10558("type"));
		return REGISTRY.method_10223(id).fromNbt(nbt);
	}

	public static AbstractPocketBuilder<?, ?> deserializeBuilder(class_2487 nbt) {
		class_2960 id = class_2960.method_12829(nbt.method_10558("type"));
		return REGISTRY.method_10223(id).builder().fromNbt(nbt);
	}

	public static class_2487 serialize(AbstractPocket<?> pocket) {
		return pocket.toNbt(new class_2487());
	}

	public V fromNbt(class_2487 nbt) {
		this.id = nbt.method_10550("id");
		this.world = class_5321.method_29179(class_2378.field_25298, new class_2960(nbt.method_10558("world")));

		return (V) this;
	}

	public class_2487 toNbt(class_2487 nbt) {
		nbt.method_10569("id", id);
		nbt.method_10582("world", world.method_29177().toString());

		getType().toNbt(nbt);

		return nbt;
	}

	public abstract AbstractPocketType<?> getType();

	public Map<String, Double> toVariableMap(Map<String, Double> variableMap) {
		variableMap.put("id", (double) this.id);
		return variableMap;
	}

	public abstract Pocket getReferencedPocket();

	// for bypassing the world check in some cases
	public Pocket getReferencedPocket(PocketDirectory directory) {
		return getReferencedPocket();
	}

	public class_5321<class_1937> getWorld() {
		return world;
	}

	public interface AbstractPocketType<T extends AbstractPocket<?>> {
		AbstractPocketType<IdReferencePocket> ID_REFERENCE = register(new class_2960("dimdoors", IdReferencePocket.KEY), IdReferencePocket::new, IdReferencePocket::builder);

		AbstractPocketType<Pocket> POCKET = register(new class_2960("dimdoors", Pocket.KEY), Pocket::new, Pocket::builder);
		AbstractPocketType<PrivatePocket> PRIVATE_POCKET = register(new class_2960("dimdoors", PrivatePocket.KEY), PrivatePocket::new, PrivatePocket::builderPrivatePocket);
		AbstractPocketType<LazyGenerationPocket> LAZY_GENERATION_POCKET = register(new class_2960("dimdoors", LazyGenerationPocket.KEY), LazyGenerationPocket::new, LazyGenerationPocket::builderLazyGenerationPocket);


		T fromNbt(class_2487 nbt);

		class_2487 toNbt(class_2487 nbt);

		T instance();

		AbstractPocketBuilder<?, T> builder();

		static void register() {
			DimensionalDoorsInitializer.apiSubscribers.forEach(d -> d.registerAbstractPocketTypes(REGISTRY));
		}

		static <U extends AbstractPocket<P>, P extends AbstractPocket<P>> AbstractPocketType<U> register(class_2960 id, Supplier<U> supplier, Supplier<? extends AbstractPocketBuilder<?, U>> factorySupplier) {
			return class_2378.method_10230(REGISTRY, id, new AbstractPocketType<U>() {
				@Override
				public U fromNbt(class_2487 nbt) {
					return (U) supplier.get().fromNbt(nbt);
				}

				@Override
				public class_2487 toNbt(class_2487 nbt) {
					nbt.method_10582("type", id.toString());
					return nbt;
				}

				@Override
				public U instance() {
					return supplier.get();
				}

				@Override
				public AbstractPocketBuilder<?, U> builder() {
					return factorySupplier.get();
				}
			});
		}
	}

	public static abstract class AbstractPocketBuilder<P extends AbstractPocketBuilder<P, T>, T extends AbstractPocket<?>> {
		protected final AbstractPocketType<T> type;

		private int id;
		private class_5321<class_1937> world;

		protected AbstractPocketBuilder(AbstractPocketType<T> type) {
			this.type = type;
		}

		public class_2382 getExpectedSize() {
			return new class_2382(1, 1, 1);
		}

		public T build() {
			T instance = type.instance();

			instance.id = id;
			instance.world = world;

			return instance;
		}

		public P id(int id) {
			this.id = id;
			return getSelf();
		}

		public P world(class_5321<class_1937> world) {
			this.world = world;
			return getSelf();
		}

		public P getSelf() {
			return (P) this;
		}

		abstract public P fromNbt(class_2487 nbt);

		abstract public class_2487 toNbt(class_2487 nbt);

		/*
		public P fromTag(CompoundTag tag) {
			id = tag.getInt("id");
			world = RegistryKey.of(Registry.DIMENSION, new Identifier(tag.getString("world")));

			return getSelf();
		}

		public CompoundTag toTag(CompoundTag tag) {
			tag.putInt("id", id);
			tag.putString("world", world.getValue().toString());

			return tag;
		}
		 */
	}
}
