package org.dimdev.dimdoors.world.pocket.type.addon;

import net.minecraft.class_1297;
import net.minecraft.class_1767;
import net.minecraft.class_2338;
import net.minecraft.class_2487;
import net.minecraft.class_2588;
import net.minecraft.class_2680;
import net.minecraft.class_2960;
import net.minecraft.class_3218;
import net.minecraft.class_5250;
import org.dimdev.dimdoors.DimensionalDoorsInitializer;
import org.dimdev.dimdoors.api.util.EntityUtils;
import org.dimdev.dimdoors.block.AncientFabricBlock;
import org.dimdev.dimdoors.block.FabricBlock;
import org.dimdev.dimdoors.block.ModBlocks;
import org.dimdev.dimdoors.world.pocket.type.Pocket;
import org.dimdev.dimdoors.world.pocket.type.PocketColor;
import org.dimdev.dimdoors.world.pocket.type.PrivatePocket;

public class DyeableAddon implements PocketAddon {
	public static class_2960 ID = new class_2960("dimdoors", "dyeable");

	private static final int BLOCKS_PAINTED_PER_DYE = 1000000;

	protected PocketColor dyeColor = PocketColor.WHITE;
	private PocketColor nextDyeColor = PocketColor.NONE;
	private int count = 0;

	private static int amountOfDyeRequiredToColor(Pocket pocket) {
		int outerVolume = pocket.getBox().method_14660() * pocket.getBox().method_14663() * pocket.getBox().method_14663();
		int innerVolume = (pocket.getBox().method_14660() - 5) * (pocket.getBox().method_14663() - 5) * (pocket.getBox().method_14663() - 5);

		return Math.max((outerVolume - innerVolume) / BLOCKS_PAINTED_PER_DYE, 1);
	}

	private void repaint(Pocket pocket, class_1767 dyeColor) {
		class_3218 serverWorld = DimensionalDoorsInitializer.getWorld(pocket.getWorld());
		class_2680 innerWall = ModBlocks.fabricFromDye(dyeColor).method_9564();
		class_2680 outerWall = ModBlocks.ancientFabricFromDye(dyeColor).method_9564();

		class_2338.method_23627(pocket.getBox()).forEach(pos -> {
			if (serverWorld.method_8320(pos).method_26204() instanceof AncientFabricBlock) {
				serverWorld.method_8501(pos, outerWall);
			} else if (serverWorld.method_8320(pos).method_26204() instanceof FabricBlock) {
				serverWorld.method_8501(pos, innerWall);
			}
		});
	}

	public boolean addDye(Pocket pocket, class_1297 entity, class_1767 dyeColor) {
		PocketColor color = PocketColor.from(dyeColor);

		int maxDye = amountOfDyeRequiredToColor(pocket);

		if (this.dyeColor == color) {
			EntityUtils.chat(entity, class_5250.method_43477(new class_2588("dimdoors.pockets.dyeAlreadyAbsorbed")));
			return false;
		}

		if (this.nextDyeColor != PocketColor.NONE && this.nextDyeColor == color) {
			if (this.count + 1 > maxDye) {
				repaint(pocket, dyeColor);
				this.dyeColor = color;
				this.nextDyeColor = PocketColor.NONE;
				this.count = 0;
				EntityUtils.chat(entity, class_5250.method_43477(new class_2588("dimdoors.pocket.pocketHasBeenDyed", dyeColor)));
			} else {
				this.count++;
				EntityUtils.chat(entity, class_5250.method_43477(new class_2588("dimdoors.pocket.remainingNeededDyes", this.count, maxDye, color)));
			}
		} else {
			this.nextDyeColor = color;
			this.count = 1;
			EntityUtils.chat(entity, class_5250.method_43477(new class_2588("dimdoors.pocket.remainingNeededDyes", this.count, maxDye, color)));
		}
		return true;
	}

	@Override
	public boolean applicable(Pocket pocket) {
		return pocket instanceof PrivatePocket;
	}

	@Override
	public PocketAddon fromNbt(class_2487 nbt) {

		this.dyeColor = PocketColor.from(nbt.method_10550("dyeColor"));
		this.nextDyeColor = PocketColor.from(nbt.method_10550("nextDyeColor"));
		this.count = nbt.method_10550("count");

		return this;
	}

	@Override
	public class_2487 toNbt(class_2487 nbt) {
		PocketAddon.super.toNbt(nbt);

		nbt.method_10569("dyeColor", this.dyeColor.getId());
		nbt.method_10569("nextDyeColor", this.nextDyeColor.getId());
		nbt.method_10569("count", this.count);

		return nbt;
	}

	@Override
	public PocketAddonType<? extends PocketAddon> getType() {
		return PocketAddonType.DYEABLE_ADDON;
	}

	@Override
	public class_2960 getId() {
		return ID;
	}

	public interface DyeablePocketBuilder<T extends Pocket.PocketBuilder<T, ?>> extends PocketBuilderExtension<T> {
		default T dyeColor(PocketColor dyeColor) {

			this.<DyeableBuilderAddon>getAddon(ID).dyeColor = dyeColor;

			return getSelf();
		}
	}

	public static class DyeableBuilderAddon implements PocketBuilderAddon<DyeableAddon> {

		private PocketColor dyeColor = PocketColor.NONE;
		// TODO: add some Pocket#init so that we can have boolean shouldRepaintOnInit

		@Override
		public void apply(Pocket pocket) {
			DyeableAddon addon = new DyeableAddon();
			addon.dyeColor = dyeColor;
			pocket.addAddon(addon);
		}

		@Override
		public class_2960 getId() {
			return ID;
		}

		@Override
		public PocketBuilderAddon<DyeableAddon> fromNbt(class_2487 nbt) {
			this.dyeColor = PocketColor.from(nbt.method_10550("dye_color"));

			return this;
		}

		@Override
		public class_2487 toNbt(class_2487 nbt) {
			PocketBuilderAddon.super.toNbt(nbt);

			nbt.method_10569("dye_color", dyeColor.getId());

			return nbt;
		}

		@Override
		public PocketAddonType<DyeableAddon> getType() {
			return PocketAddonType.DYEABLE_ADDON;
		}
	}

	public interface DyeablePocket extends AddonProvider {
		default boolean addDye(class_1297 entity, class_1767 dyeColor) {
			ensureIsPocket();
			if (!this.hasAddon(ID)) {
				DyeableAddon addon = new DyeableAddon();
				this.addAddon(addon);
				return addon.addDye((Pocket) this, entity, dyeColor);
			}
			return this.<DyeableAddon>getAddon(ID).addDye((Pocket) this, entity, dyeColor);
		}
	}
}
